import logging
import os.path
import urllib.parse

from django.contrib.postgres.fields import JSONField
from django.core.exceptions import ObjectDoesNotExist
from django.db import models
from django.db.models.fields import FieldDoesNotExist
from django.db.models.fields.files import FieldFile
from django.db.models.signals import pre_save

try:
    from ..cms.internal_tags.fields import (
        TaggedRelationField,
        TaggedRelationFormField
    )
except ValueError:
    from cms.internal_tags.fields import TaggedRelationFormField, TaggedRelationField

from . import get_image_cropper, settings, utils, widgets

logger = logging.getLogger(__name__)


class AssetsFileFormField(TaggedRelationFormField):
    widget = widgets.AssetsFileWidget

    def __init__(self, **kwargs):
        # Type/Tags
        self.asset_type = kwargs.pop("asset_type", None)
        self.sizes = kwargs.pop("sizes", None)
        super().__init__(**kwargs)

    def widget_attrs(self, widget):
        widget.required = self.required
        widget.asset_type = self.asset_type
        widget.sizes = self.sizes
        return {}


class AssetFieldFile(FieldFile):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        sizes = list(getattr(self.field, "image_sizes", []))
        sizes.extend(get_image_cropper().required_crops())

        for size in sizes:
            setattr(
                self, "{0}_url".format(size), utils.partial(self._get_url, version=size)
            )

    def get_version(self, version, cbversion=None):
        return self._get_url(version, cbversion=cbversion)

    def _get_url(self, version=None, cbversion=None):
        if not self:
            return ""
        url = super().url
        if url:
            if version:
                url_parts = list(urllib.parse.urlsplit(url))
                _, extension = os.path.splitext(url_parts[2])
                if extension[1:] in utils.IMAGE_EXTENSIONS:
                    url_parts[2] = utils.get_size_filename(url_parts[2], version)
                    url = urllib.parse.urlunsplit(url_parts)

            if settings.USE_CACHE_BUST:
                if not cbversion:
                    if getattr(self.instance, "cbversion", None):
                        cbversion = getattr(self.instance, "cbversion")

                if not cbversion:
                    cbversion = utils.get_cache_bust_version(url)

        if cbversion:
            url = "{0}?v={1}".format(url, cbversion)
        return url

    url = property(_get_url)


class AssetRealFileField(models.FileField):
    attr_class = AssetFieldFile

    def __init__(self, *args, **kwargs):
        self.image_sizes = kwargs.pop("image_sizes", [])
        super().__init__(*args, **kwargs)


class AssetsFileField(TaggedRelationField):
    default_form_class = AssetsFileFormField
    default_cache_field_class = JSONField

    def __init__(self, *args, **kwargs):
        if "related_name" not in kwargs:
            kwargs["related_name"] = "+"

        if "on_delete" not in kwargs:
            kwargs["on_delete"] = models.PROTECT

        self.cache_field_class = kwargs.pop(
            "cache_field_class", self.default_cache_field_class
        )
        self.asset_type = kwargs.pop("type", "unknown")
        self.denormalize = kwargs.pop("denormalize", True)

        image_sizes = kwargs.pop("image_sizes", [])
        cropper = get_image_cropper()

        self.image_sizes = []
        from .crops import CropConfig

        if isinstance(image_sizes, dict):
            for k, v in list(image_sizes.items()):
                if v and isinstance(v, dict):
                    cropper.register(CropConfig(k, **v))
                    self.image_sizes.append(k)
        else:
            for c in image_sizes:
                cropper.register(c)
                self.image_sizes.append(c.name)

        kwargs["to"] = settings.ASSET_MODEL
        return super().__init__(**kwargs)

    def get_formfield_defaults(self):
        # This is a fairly standard way to set up some defaults
        # while letting the caller override them.
        defaults = super().get_formfield_defaults()
        defaults["asset_type"] = self.asset_type
        defaults["sizes"] = self.image_sizes
        return defaults

    def contribute_to_class(self, cls, name):
        """
        `get_field` and exception only work if the asset field was declared
        later than other fields since they are created from top to bottom.
        For example if we have model with these fields:
            image = AssetsFileField()
            image_alt_text = CharField()
            image_urls = JSONField()
        If we run cls._meta.get_field("image_alt_text"/"image_urls")
        during creation of `image` field - we'll get FieldDoesNotExist error
        even though the field was declared.
        """
        if self.denormalize and not cls._meta.abstract:
            cache_name = self.get_denormalized_field_name(name)

            try:
                cls._meta.get_field(cache_name)
            except FieldDoesNotExist:
                denormalize_field = self.cache_field_class(
                    default=dict, editable=False,
                )

                cls.add_to_class(cache_name, denormalize_field)

            if "image" in self.asset_type:
                alt_name = self.get_alt_text_field_name(name)
                try:
                    cls._meta.get_field(alt_name)
                except FieldDoesNotExist:
                    denormalize_field = models.CharField(
                        max_length=255, blank=True, null=True,
                    )
                    cls.add_to_class(alt_name, denormalize_field)

            pre_save.connect(denormalize_assets, sender=cls)

        # add the field normally
        super().contribute_to_class(cls, name)

    def get_alt_text_field_name(self, name):
        return "{}_alt_text".format(name)

    def get_denormalized_field_name(self, name):
        return "{0}_urls".format(name)

    def deconstruct(self):
        """
        Denormalize is always false migrations
        """
        name, path, args, kwargs = super().deconstruct()
        kwargs["denormalize"] = False
        return name, path, args, kwargs


def denormalize_assets(sender, instance, **kwargs):
    for field in instance._meta.fields:
        if isinstance(field, AssetsFileField):
            cache_name = field.get_denormalized_field_name(field.name)
            try:
                asset_ins = getattr(instance, field.name)
                if field.denormalize:
                    if asset_ins:
                        setattr(
                            instance, cache_name,
                            asset_ins.create_urls_json(field.image_sizes)
                        )
                        asset_ins.ensure_crops(*field.image_sizes)
                    else:
                        setattr(instance, cache_name, "")
            except ObjectDoesNotExist:
                setattr(instance, cache_name, "")



# Licensed under the Apache License: http://www.apache.org/licenses/LICENSE-2.0
# For details: https://github.com/nedbat/coveragepy/blob/master/NOTICE.txt

"""Results of coverage measurement."""

import collections

from coverage.backward import iitems
from coverage.debug import SimpleReprMixin
from coverage.misc import contract, CoverageException, nice_pair


class Analysis(object):
    """The results of analyzing a FileReporter."""

    def __init__(self, data, file_reporter, file_mapper):
        self.data = data
        self.file_reporter = file_reporter
        self.filename = file_mapper(self.file_reporter.filename)
        self.statements = self.file_reporter.lines()
        self.excluded = self.file_reporter.excluded_lines()

        # Identify missing statements.
        executed = self.data.lines(self.filename) or []
        executed = self.file_reporter.translate_lines(executed)
        self.executed = executed
        self.missing = self.statements - self.executed

        if self.data.has_arcs():
            self._arc_possibilities = sorted(self.file_reporter.arcs())
            self.exit_counts = self.file_reporter.exit_counts()
            self.no_branch = self.file_reporter.no_branch_lines()
            n_branches = self._total_branches()
            mba = self.missing_branch_arcs()
            n_partial_branches = sum(len(v) for k,v in iitems(mba) if k not in self.missing)
            n_missing_branches = sum(len(v) for k,v in iitems(mba))
        else:
            self._arc_possibilities = []
            self.exit_counts = {}
            self.no_branch = set()
            n_branches = n_partial_branches = n_missing_branches = 0

        self.numbers = Numbers(
            n_files=1,
            n_statements=len(self.statements),
            n_excluded=len(self.excluded),
            n_missing=len(self.missing),
            n_branches=n_branches,
            n_partial_branches=n_partial_branches,
            n_missing_branches=n_missing_branches,
        )

    def missing_formatted(self, branches=False):
        """The missing line numbers, formatted nicely.

        Returns a string like "1-2, 5-11, 13-14".

        If `branches` is true, includes the missing branch arcs also.

        """
        if branches and self.has_arcs():
            arcs = iitems(self.missing_branch_arcs())
        else:
            arcs = None

        return format_lines(self.statements, self.missing, arcs=arcs)

    def has_arcs(self):
        """Were arcs measured in this result?"""
        return self.data.has_arcs()

    @contract(returns='list(tuple(int, int))')
    def arc_possibilities(self):
        """Returns a sorted list of the arcs in the code."""
        return self._arc_possibilities

    @contract(returns='list(tuple(int, int))')
    def arcs_executed(self):
        """Returns a sorted list of the arcs actually executed in the code."""
        executed = self.data.arcs(self.filename) or []
        executed = self.file_reporter.translate_arcs(executed)
        return sorted(executed)

    @contract(returns='list(tuple(int, int))')
    def arcs_missing(self):
        """Returns a sorted list of the arcs in the code not executed."""
        possible = self.arc_possibilities()
        executed = self.arcs_executed()
        missing = (
            p for p in possible
                if p not in executed
                    and p[0] not in self.no_branch
        )
        return sorted(missing)

    @contract(returns='list(tuple(int, int))')
    def arcs_unpredicted(self):
        """Returns a sorted list of the executed arcs missing from the code."""
        possible = self.arc_possibilities()
        executed = self.arcs_executed()
        # Exclude arcs here which connect a line to itself.  They can occur
        # in executed data in some cases.  This is where they can cause
        # trouble, and here is where it's the least burden to remove them.
        # Also, generators can somehow cause arcs from "enter" to "exit", so
        # make sure we have at least one positive value.
        unpredicted = (
            e for e in executed
                if e not in possible
                    and e[0] != e[1]
                    and (e[0] > 0 or e[1] > 0)
        )
        return sorted(unpredicted)

    def _branch_lines(self):
        """Returns a list of line numbers that have more than one exit."""
        return [l1 for l1,count in iitems(self.exit_counts) if count > 1]

    def _total_branches(self):
        """How many total branches are there?"""
        return sum(count for count in self.exit_counts.values() if count > 1)

    @contract(returns='dict(int: list(int))')
    def missing_branch_arcs(self):
        """Return arcs that weren't executed from branch lines.

        Returns {l1:[l2a,l2b,...], ...}

        """
        missing = self.arcs_missing()
        branch_lines = set(self._branch_lines())
        mba = collections.defaultdict(list)
        for l1, l2 in missing:
            if l1 in branch_lines:
                mba[l1].append(l2)
        return mba

    @contract(returns='dict(int: tuple(int, int))')
    def branch_stats(self):
        """Get stats about branches.

        Returns a dict mapping line numbers to a tuple:
        (total_exits, taken_exits).
        """

        missing_arcs = self.missing_branch_arcs()
        stats = {}
        for lnum in self._branch_lines():
            exits = self.exit_counts[lnum]
            try:
                missing = len(missing_arcs[lnum])
            except KeyError:
                missing = 0
            stats[lnum] = (exits, exits - missing)
        return stats


class Numbers(SimpleReprMixin):
    """The numerical results of measuring coverage.

    This holds the basic statistics from `Analysis`, and is used to roll
    up statistics across files.

    """
    # A global to determine the precision on coverage percentages, the number
    # of decimal places.
    _precision = 0
    _near0 = 1.0              # These will change when _precision is changed.
    _near100 = 99.0

    def __init__(self, n_files=0, n_statements=0, n_excluded=0, n_missing=0,
                    n_branches=0, n_partial_branches=0, n_missing_branches=0
                    ):
        self.n_files = n_files
        self.n_statements = n_statements
        self.n_excluded = n_excluded
        self.n_missing = n_missing
        self.n_branches = n_branches
        self.n_partial_branches = n_partial_branches
        self.n_missing_branches = n_missing_branches

    def init_args(self):
        """Return a list for __init__(*args) to recreate this object."""
        return [
            self.n_files, self.n_statements, self.n_excluded, self.n_missing,
            self.n_branches, self.n_partial_branches, self.n_missing_branches,
        ]

    @classmethod
    def set_precision(cls, precision):
        """Set the number of decimal places used to report percentages."""
        assert 0 <= precision < 10
        cls._precision = precision
        cls._near0 = 1.0 / 10**precision
        cls._near100 = 100.0 - cls._near0

    @property
    def n_executed(self):
        """Returns the number of executed statements."""
        return self.n_statements - self.n_missing

    @property
    def n_executed_branches(self):
        """Returns the number of executed branches."""
        return self.n_branches - self.n_missing_branches

    @property
    def pc_covered(self):
        """Returns a single percentage value for coverage."""
        if self.n_statements > 0:
            numerator, denominator = self.ratio_covered
            pc_cov = (100.0 * numerator) / denominator
        else:
            pc_cov = 100.0
        return pc_cov

    @property
    def pc_covered_str(self):
        """Returns the percent covered, as a string, without a percent sign.

        Note that "0" is only returned when the value is truly zero, and "100"
        is only returned when the value is truly 100.  Rounding can never
        result in either "0" or "100".

        """
        pc = self.pc_covered
        if 0 < pc < self._near0:
            pc = self._near0
        elif self._near100 < pc < 100:
            pc = self._near100
        else:
            pc = round(pc, self._precision)
        return "%.*f" % (self._precision, pc)

    @classmethod
    def pc_str_width(cls):
        """How many characters wide can pc_covered_str be?"""
        width = 3   # "100"
        if cls._precision > 0:
            width += 1 + cls._precision
        return width

    @property
    def ratio_covered(self):
        """Return a numerator and denominator for the coverage ratio."""
        numerator = self.n_executed + self.n_executed_branches
        denominator = self.n_statements + self.n_branches
        return numerator, denominator

    def __add__(self, other):
        nums = Numbers()
        nums.n_files = self.n_files + other.n_files
        nums.n_statements = self.n_statements + other.n_statements
        nums.n_excluded = self.n_excluded + other.n_excluded
        nums.n_missing = self.n_missing + other.n_missing
        nums.n_branches = self.n_branches + other.n_branches
        nums.n_partial_branches = (
            self.n_partial_branches + other.n_partial_branches
            )
        nums.n_missing_branches = (
            self.n_missing_branches + other.n_missing_branches
            )
        return nums

    def __radd__(self, other):
        # Implementing 0+Numbers allows us to sum() a list of Numbers.
        if other == 0:
            return self
        return NotImplemented


def _line_ranges(statements, lines):
    """Produce a list of ranges for `format_lines`."""
    statements = sorted(statements)
    lines = sorted(lines)

    pairs = []
    start = None
    lidx = 0
    for stmt in statements:
        if lidx >= len(lines):
            break
        if stmt == lines[lidx]:
            lidx += 1
            if not start:
                start = stmt
            end = stmt
        elif start:
            pairs.append((start, end))
            start = None
    if start:
        pairs.append((start, end))
    return pairs


def format_lines(statements, lines, arcs=None):
    """Nicely format a list of line numbers.

    Format a list of line numbers for printing by coalescing groups of lines as
    long as the lines represent consecutive statements.  This will coalesce
    even if there are gaps between statements.

    For example, if `statements` is [1,2,3,4,5,10,11,12,13,14] and
    `lines` is [1,2,5,10,11,13,14] then the result will be "1-2, 5-11, 13-14".

    Both `lines` and `statements` can be any iterable. All of the elements of
    `lines` must be in `statements`, and all of the values must be positive
    integers.

    If `arcs` is provided, they are (start,[end,end,end]) pairs that will be
    included in the output as long as start isn't in `lines`.

    """
    line_items = [(pair[0], nice_pair(pair)) for pair in _line_ranges(statements, lines)]
    if arcs:
        line_exits = sorted(arcs)
        for line, exits in line_exits:
            for ex in sorted(exits):
                if line not in lines:
                    dest = (ex if ex > 0 else "exit")
                    line_items.append((line, "%d->%s" % (line, dest)))

    ret = ', '.join(t[-1] for t in sorted(line_items))
    return ret


@contract(total='number', fail_under='number', precision=int, returns=bool)
def should_fail_under(total, fail_under, precision):
    """Determine if a total should fail due to fail-under.

    `total` is a float, the coverage measurement total. `fail_under` is the
    fail_under setting to compare with. `precision` is the number of digits
    to consider after the decimal point.

    Returns True if the total should fail.

    """
    # We can never achieve higher than 100% coverage, or less than zero.
    if not (0 <= fail_under <= 100.0):
        msg = "fail_under={} is invalid. Must be between 0 and 100.".format(fail_under)
        raise CoverageException(msg)

    # Special case for fail_under=100, it must really be 100.
    if fail_under == 100.0 and total != 100.0:
        return True

    return round(total, precision) < fail_under



from urllib.parse import urlencode

from django import http
from django.db import models
from django.forms import models as model_forms
from django.views.generic.list import MultipleObjectMixin

from . import fields, helpers, renders, transaction, widgets
from .actions import CHECKBOX_NAME
from .base_views import ModelCMSMixin, ModelCMSView
from .forms import VersionFilterForm
from .models import CMSLog


class ListView(ModelCMSMixin, MultipleObjectMixin, ModelCMSView):
    """
    Class for listing objects and optionally editing those objects.
    Inherits from ModelCMSMixin, MultipleObjectMixin, ModelCMSView.
    This class adds a ChoicesRender instance as a 'choices' render
    type.

    :param display_fields: List of fields to display in the table. \
    Defaults to ('__str__',).
    :param change_fields: List of fields that are editable in this \
    view. Defaults to empty tuple.
    :param form_class: A model form class to use.
    :param filter_form: An instance of BaseFilterForm that can return \
    filter parameters that can be used to filer this list. \
    Defaults to None.
    :param default_template: Defaults to 'cms/list.html'.
    :param can_sort: Does this view allow sorting by column. \
    True or False.
    :param action_links: A list of action links to display \
    with each row. See `get_actions` for details.
    :param origin_action_links: A list of action links that should \
    redirect back here preserving the querystring.
    :param paginate_by: How many items per page. Defaults \
    to None, which means no paginating.
    :param paginator_class: The paginator class to use. \
    Defaults to Paginator.
    :param parent_preview: Should we show 'Preview' button \
    that links to parent object. \
    Defaults to False.
    """

    # Fields that can be edited in this list.
    change_fields = ()

    # Fields that should be displayed. Must include
    display_fields = ("__str__",)

    # class that each form should use
    form_class = None

    filter_form = None

    action_views = None

    default_template = "cms/list.html"

    # Are we allowed to sort this list?
    can_sort = True

    # Preview button with parent's URL to preview
    parent_preview = False

    action_links = (
        ("edit", "Edit", "e"),
        ("delete", "Delete", "d", True),
        ("publish", "Publish", "P", True),
        ("preview", "Preview", "p"),
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.renders["choices"] = renders.ChoicesRender()

    def get_back_bundle(self, start_bundle=None):
        return None

    def formfield_for_dbfield(self, db_field, **kwargs):
        """
        Same as parent but sets the widget for any OrderFields to
        HiddenTextInput.
        """
        if isinstance(db_field, fields.OrderField):
            kwargs["widget"] = widgets.HiddenTextInput

        return super().formfield_for_dbfield(db_field, **kwargs)

    def _verify_list(self):
        allow_empty = self.get_allow_empty()
        if not allow_empty and len(self.object_list) == 0:
            raise http.Http404

    def get_actions(self):
        """
        Hook for specifying action links.
        Returns `self.action_links` by default.

        The format is a list of tuples. Each tuple
        has four slots:

        * **0** The view name to lookup to get the link.
        * **1** The verbose name of the link.
        * **2** The css attribute for this link.
        * **3** Should this link include an origin redirect
        """

        return self.action_links

    def get_filter_form(self, **kwargs):
        """
        If there is a filter_form, initializes that
        form with the contents of request.GET and
        returns it.
        """

        form = None
        if self.filter_form:
            form = self.filter_form(self.request.GET)
        elif self.model and hasattr(self.model._meta, "_is_view"):
            form = VersionFilterForm(self.request.GET)
        return form

    def get_filter(self, **filter_kwargs):
        """
        Combines the Q objects returned by a valid
        filter form with any other arguments and
        returns a list of Q objects that can be passed
        to a queryset.
        """

        q_objects = super().get_filter(**filter_kwargs)
        form = self.get_filter_form()
        if form:
            q_objects.extend(form.get_filter())

        return q_objects

    def get_formset_form_class(self):
        """
        Returns the form class for use in the formset.

        If a form_class attribute or change_fields
        is provided then a form will be constructed
        with that. Otherwise None is returned.
        """
        if self.form_class or self.change_fields:
            params = {"formfield_callback": self.formfield_for_dbfield}
            if self.form_class:
                fc = self.customize_form_widgets(self.form_class)
                params["form"] = fc
            if self.change_fields:
                params["fields"] = self.change_fields

            return model_forms.modelform_factory(self.model, **params)

    def get_formset_class(self, **kwargs):
        """
        Returns the formset for the queryset,
        if a form class is available.
        """
        form_class = self.get_formset_form_class()
        if form_class:
            kwargs["formfield_callback"] = self.formfield_for_dbfield
            return model_forms.modelformset_factory(
                self.model, form_class, fields=self.change_fields, extra=0, **kwargs
            )

    def _add_formset_id(self, data, queryset):
        q = None
        key = self.model._meta.pk.name
        for k, v in list(data.items()):
            if k.endswith(f"-{key}"):
                new_q = models.Q(**{key: v})
                if q:
                    q = q | new_q
                else:
                    q = new_q
        queryset = queryset.filter(q)
        return queryset

    def get_formset(self, data=None, queryset=None):
        """
        Returns an instantiated FormSet if available.
        If `self.can_submit` is False then no formset
        is returned.
        """
        if not self.can_submit:
            return None

        FormSet = self.get_formset_class()
        if queryset is None:
            queryset = self.get_queryset()

        if FormSet:
            if data:
                queryset = self._add_formset_id(data, queryset)
            return FormSet(data, queryset=queryset)

    def get_visible_fields(self, formset):
        """
        Returns a list of visible fields. This
        are all the fields in `self.display_fields`
        plus any visible fields in the given formset
        minus any hidden fields in the formset.
        """

        visible_fields = list(self.display_fields)
        if formset:
            for x in formset.empty_form.visible_fields():
                if x.name not in visible_fields:
                    visible_fields.append(x.name)

            for x in formset.empty_form.hidden_fields():
                if x.name in visible_fields:
                    visible_fields.remove(x.name)

        return visible_fields

    def _get_query_string(self, request, exclude_page=True):
        qs = request.GET.copy()
        if "page" in qs and exclude_page:
            qs.pop("page")

        return f"?{qs.urlencode()}&" if qs else "?"

    def _sort_queryset(self, object_list, sort_field, order_type):
        sort_field = helpers.get_sort_field(sort_field, object_list.model)

        if order_type and sort_field:
            if order_type == helpers.AdminList.ASC:
                queryset = self.object_list.order_by(sort_field)
            else:
                queryset = self.object_list.order_by("-" + sort_field)
        else:
            queryset = self.object_list
            if not queryset.ordered:
                queryset = self.object_list.order_by("-pk")

        return queryset

    def _paginate_queryset(self, queryset):
        page_size = self.get_paginate_by(queryset)
        paginator = None
        if page_size:
            paginator, page, queryset, is_paginated = self.paginate_queryset(
                queryset, page_size
            )
            is_paginated = True
        else:
            page = None
            is_paginated = False
        return is_paginated, page, paginator, queryset

    def get_action_context(self, request):
        actions = []

        for action in self.action_views or self.bundle._meta.action_views:
            action_name = "{0}{1}".format(action, self.bundle.action_alias)
            v, name = self.bundle.get_initialized_view_and_name(
                action_name, **self.kwargs
            )
            url = self.bundle.get_view_url(
                action_name, request.user, url_kwargs=self.kwargs
            )
            if v and url:
                actions.append((url, v.action_name))

        return actions

    def get_list_data(self, request, **kwargs):
        """
        Returns the data needed for displaying the list.
        Returns a dictionary that should be treating as
        context with the following arguments:

        * **list** - The list of data to be displayed. This is \
        an instance of a wrapper class that combines the queryset \
        and formset and provides looping and labels and sorting controls.
        * **filter_form** - An instance of your filter form.
        * **is_paginated** - Is this list paginated, True or False.
        * **paginator** - The paginator object if available.
        * **page_obj** - A pagination object if available.
        * **show_form** - Should the rendering template show form controls.
        * **parent_preview_url** - URL to preview parent object.
        """

        self.object_list = self.get_queryset()
        self._verify_list()

        sort_field = None
        order_type = None
        if self.can_sort:
            default = None
            default_order = helpers.AdminList.ASC
            if self.object_list.ordered:
                if self.object_list.query.order_by:
                    default = self.object_list.query.order_by[0]
                else:
                    default = self.object_list.model._meta.ordering[0]
                if default.startswith("-"):
                    default = default[1:]
                    default_order = helpers.AdminList.DESC

            sort_field = request.GET.get("sf", default)
            order_type = request.GET.get("ot", default_order)

        queryset = self._sort_queryset(self.object_list, sort_field, order_type)

        if self.request.method == "POST" and self.can_submit:
            formset = self.get_formset(data=self.request.POST, queryset=queryset)
            is_paginated, page, paginator, queryset = self._paginate_queryset(queryset)
        else:
            is_paginated, page, paginator, queryset = self._paginate_queryset(queryset)
            formset = self.get_formset(queryset=queryset)

        visible_fields = self.get_visible_fields(formset)
        adm_list = helpers.AdminList(
            formset, queryset, visible_fields, sort_field, order_type, self.model_name
        )

        parent_preview_url = None
        if self.parent_preview:
            parent_preview_url = self.parent_object.get_absolute_url()

        actions = self.get_action_context(request)
        data = {
            "list": adm_list,
            "filter_form": self.get_filter_form(),
            "page_obj": page,
            "is_paginated": is_paginated,
            "show_form": (self.can_submit and formset is not None),
            "paginator": paginator,
            "checkbox_name": CHECKBOX_NAME,
            "actions": actions,
            "parent_preview_url": parent_preview_url,
        }

        return data

    def get_context_data(self, **kwargs):
        """
        Get the context for this view. Adds the following values:

        * **query_string** - The querystring minus the current page.
        * **action_links** - The results of the `get_actions` method.
        """

        origin_qs = self._get_query_string(self.request, False)
        context = {
            "query_string": self._get_query_string(self.request),
            "origin_qs": self.request.path + origin_qs,
            "origin_var": self.ORIGIN_ARGUMENT,
            "action_links": self.get_actions(),
        }
        context.update(kwargs)

        return context

    def get(self, request, *args, **kwargs):
        """
        Method for handling GET requests. If there is
        a GET parameter type=choice, then the render_type
        will be set to 'choices' to return a JSON version
        of this list. Calls `render` with the data from the
        `get_list_data` method as context.
        """

        if request.GET.get("type") == "choices":
            self.render_type = "choices"
            self.can_submit = False

        data = self.get_list_data(request, **kwargs)
        return self.render(request, **data)

    def post(self, request, *args, **kwargs):
        """
        Method for handling POST requests.
        If the formset is valid this will
        loop through the formset and save each form.
        A log is generated for each save. The user
        is notified of the total number of changes
        with a message. Returns a 'render redirect' to
        the current url.

        TODO: These formsets suffer from the same potential concurrency
        issues that the django admin has. This is caused by some issues
        with django formsets and concurrent users editing the same
        objects.
        """
        msg = None
        action = request.POST.get("actions", None)
        selected = request.POST.getlist(CHECKBOX_NAME)
        if not action == "None" and action is not None:
            if len(selected) > 0:
                sel = {CHECKBOX_NAME: ",".join(selected)}
                qs = "?" + urlencode(sel)
                return self.render(request, redirect_url=action + qs)

        data = self.get_list_data(request, **kwargs)

        l = data.get("list")
        formset = None
        if l and l.formset:
            formset = l.formset

        url = self.request.build_absolute_uri()
        if formset:
            # Normally calling validate on a formset.
            # will result in a db call for each pk in
            # the formset regardless if the form has
            # changed or not.
            # To try to reduce queries only do a full
            # validate on forms that changed.
            # TODO: Find a way to not have to do
            # a pk lookup for any since we already
            # have the instance we want
            for form in formset.forms:
                if not form.has_changed():
                    form.cleaned_data = {}
                    form._errors = {}

        if formset and formset.is_valid():
            changecount = 0
            with transaction.commit_on_success():
                for form in formset.forms:
                    if form.has_changed():
                        obj = form.save()
                        changecount += 1
                        self.log_action(
                            obj, CMSLog.SAVE, url=url, update_parent=changecount == 1
                        )

            return self.render(
                request,
                redirect_url=url,
                message=f"{changecount} items updated",
                collect_render_data=False,
            )
        else:
            return self.render(request, message=msg, **data)



# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the BSD License. See the LICENSE file in the root of this repository
# for complete details.
from __future__ import absolute_import, division, print_function

import re

from ._typing import MYPY_CHECK_RUNNING
from .version import InvalidVersion, Version

if MYPY_CHECK_RUNNING:  # pragma: no cover
    from typing import Union

_canonicalize_regex = re.compile(r"[-_.]+")


def canonicalize_name(name):
    # type: (str) -> str
    # This is taken from PEP 503.
    return _canonicalize_regex.sub("-", name).lower()


def canonicalize_version(_version):
    # type: (str) -> Union[Version, str]
    """
    This is very similar to Version.__str__, but has one subtle difference
    with the way it handles the release segment.
    """

    try:
        version = Version(_version)
    except InvalidVersion:
        # Legacy versions cannot be normalized
        return _version

    parts = []

    # Epoch
    if version.epoch != 0:
        parts.append("{0}!".format(version.epoch))

    # Release segment
    # NB: This strips trailing '.0's to normalize
    parts.append(re.sub(r"(\.0)+$", "", ".".join(str(x) for x in version.release)))

    # Pre-release
    if version.pre is not None:
        parts.append("".join(str(x) for x in version.pre))

    # Post-release
    if version.post is not None:
        parts.append(".post{0}".format(version.post))

    # Development release
    if version.dev is not None:
        parts.append(".dev{0}".format(version.dev))

    # Local version segment
    if version.local is not None:
        parts.append("+{0}".format(version.local))

    return "".join(parts)



import json
import logging
from typing import Optional, List, Tuple

from django.conf import settings
import requests

BASE_URL = "https://api.cloudflare.com/client/v4"
HEADERS = {
    "Authorization": f"Bearer {settings.CLOUDFLARE['AUTHORIZATION']}",
    "Content-Type": "application/json",
}


def purge_remote_cache(urls: Optional[List[str]] = None) -> Tuple[bool, str]:
    """
    Function to purge saved pages on remote cache.
    At the moment only Cloudflare is supported.
    """
    api_endpoint = f"{BASE_URL}/zones/{settings.CLOUDFLARE['ZONE_ID']}/purge_cache"
    data = (
        json.dumps({"files": urls}) if urls else json.dumps({"purge_everything": True})
    )
    r = requests.post(url=api_endpoint, data=data, headers=HEADERS, timeout=6.0)
    log_func = logging.info if r.ok else logging.error
    log_func(r.json())

    if r.ok:
        return True, "Cache flushed successfully"
    else:
        return False, f"Something went wrong. Response: {r.json()}"



import logging

from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.utils import timezone

from . import fields

logger = logging.getLogger(__name__)


class Schedulable(models.Model):
    """
    Abstract model that should be implemented
    by models that need to be scheduled.
    """

    class Meta:
        abstract = True

    def get_scheduled_filter_args(self):
        """
        Hook to provide the arguments to identify
        the object being operated on.
        """

        return {"pk": self.pk}

    def schedule(self, when=None, action=None, **kwargs):
        """
        Schedule an update of this object.

        when: The date for the update.

        action: if provided it will be looked up
        on the implementing class and called with
        **kwargs. If action is not provided each k/v pair
        in kwargs will be set on self and then self
        is saved.

        kwargs: any other arguments you would like passed
        for this change. Saved as a json object so must cleanly
        serialize.
        """

        # when is empty, just save it now.
        if not when:
            self.do_scheduled_update(action, **kwargs)
        else:
            ctype = ContentType.objects.get_for_model(self.__class__)
            Schedule(
                content_type=ctype,
                object_args=self.get_scheduled_filter_args(),
                when=when,
                action=action,
                json_args=kwargs,
            ).save()

    def do_scheduled_update(self, action, **kwargs):
        """
        Do the actual update.

        action: if provided it will be looked up
        on the implementing class and called with
        **kwargs. If action is not provided each k/v pair
        in kwargs will be set on self and then self
        is saved.

        kwargs: any other you passed for this update
        passed along to whichever method performs
        the update.
        """

        action = getattr(self, action, None)
        if callable(action):
            return action(**kwargs)
        else:
            for k, v in list(kwargs.items()):
                setattr(self, k, v)
            self.save()


class Schedule(models.Model):
    """
    Model to store scheduled updates.
    """

    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)
    object_args = fields.JSONField()

    when = models.DateTimeField()
    action = models.CharField(max_length=255, null=True)
    json_args = fields.JSONField()

    def do_updates(self):
        # Only run if we are ready
        if self.when <= timezone.now():
            klass = self.content_type.model_class()
            for obj in klass.objects.filter(**self.object_args):
                obj.do_scheduled_update(
                    self.action, date_published=self.when, **self.json_args
                )
        self.delete()

    class Meta:
        app_label = "scheduling"



import importlib

from .. import settings

if settings.INTERNAL_TAG_HANDLER.startswith("."):
    handler = importlib.import_module(__name__ + settings.INTERNAL_TAG_HANDLER)
else:
    handler = importlib.import_module(settings.INTERNAL_TAG_HANDLER)



# Copyright (c) 2008-2009 Aryeh Leib Taurog, all rights reserved.
# Released under the New BSD license.
"""
This module contains a base type which provides list-style mutations
without specific data storage methods.

See also http://static.aryehleib.com/oldsite/MutableLists.html

Author: Aryeh Leib Taurog.
"""
from functools import total_ordering


@total_ordering
class ListMixin:
    """
    A base class which provides complete list interface.
    Derived classes must call ListMixin's __init__() function
    and implement the following:

    function _get_single_external(self, i):
        Return single item with index i for general use.
        The index i will always satisfy 0 <= i < len(self).

    function _get_single_internal(self, i):
        Same as above, but for use within the class [Optional]
        Note that if _get_single_internal and _get_single_internal return
        different types of objects, _set_list must distinguish
        between the two and handle each appropriately.

    function _set_list(self, length, items):
        Recreate the entire object.

        NOTE: items may be a generator which calls _get_single_internal.
        Therefore, it is necessary to cache the values in a temporary:
            temp = list(items)
        before clobbering the original storage.

    function _set_single(self, i, value):
        Set the single item at index i to value [Optional]
        If left undefined, all mutations will result in rebuilding
        the object using _set_list.

    function __len__(self):
        Return the length

    int _minlength:
        The minimum legal length [Optional]

    int _maxlength:
        The maximum legal length [Optional]

    type or tuple _allowed:
        A type or tuple of allowed item types [Optional]
    """

    _minlength = 0
    _maxlength = None

    # ### Python initialization and special list interface methods ###

    def __init__(self, *args, **kwargs):
        if not hasattr(self, '_get_single_internal'):
            self._get_single_internal = self._get_single_external

        if not hasattr(self, '_set_single'):
            self._set_single = self._set_single_rebuild
            self._assign_extended_slice = self._assign_extended_slice_rebuild

        super().__init__(*args, **kwargs)

    def __getitem__(self, index):
        "Get the item(s) at the specified index/slice."
        if isinstance(index, slice):
            return [self._get_single_external(i) for i in range(*index.indices(len(self)))]
        else:
            index = self._checkindex(index)
            return self._get_single_external(index)

    def __delitem__(self, index):
        "Delete the item(s) at the specified index/slice."
        if not isinstance(index, (int, slice)):
            raise TypeError("%s is not a legal index" % index)

        # calculate new length and dimensions
        origLen = len(self)
        if isinstance(index, int):
            index = self._checkindex(index)
            indexRange = [index]
        else:
            indexRange = range(*index.indices(origLen))

        newLen = origLen - len(indexRange)
        newItems = (self._get_single_internal(i)
                    for i in range(origLen)
                    if i not in indexRange)

        self._rebuild(newLen, newItems)

    def __setitem__(self, index, val):
        "Set the item(s) at the specified index/slice."
        if isinstance(index, slice):
            self._set_slice(index, val)
        else:
            index = self._checkindex(index)
            self._check_allowed((val,))
            self._set_single(index, val)

    # ### Special methods for arithmetic operations ###
    def __add__(self, other):
        'add another list-like object'
        return self.__class__([*self, *other])

    def __radd__(self, other):
        'add to another list-like object'
        return other.__class__([*other, *self])

    def __iadd__(self, other):
        'add another list-like object to self'
        self.extend(other)
        return self

    def __mul__(self, n):
        'multiply'
        return self.__class__(list(self) * n)

    def __rmul__(self, n):
        'multiply'
        return self.__class__(list(self) * n)

    def __imul__(self, n):
        'multiply'
        if n <= 0:
            del self[:]
        else:
            cache = list(self)
            for i in range(n - 1):
                self.extend(cache)
        return self

    def __eq__(self, other):
        olen = len(other)
        for i in range(olen):
            try:
                c = self[i] == other[i]
            except IndexError:
                # self must be shorter
                return False
            if not c:
                return False
        return len(self) == olen

    def __lt__(self, other):
        olen = len(other)
        for i in range(olen):
            try:
                c = self[i] < other[i]
            except IndexError:
                # self must be shorter
                return True
            if c:
                return c
            elif other[i] < self[i]:
                return False
        return len(self) < olen

    # ### Public list interface Methods ###
    # ## Non-mutating ##
    def count(self, val):
        "Standard list count method"
        count = 0
        for i in self:
            if val == i:
                count += 1
        return count

    def index(self, val):
        "Standard list index method"
        for i in range(0, len(self)):
            if self[i] == val:
                return i
        raise ValueError('%s not found in object' % val)

    # ## Mutating ##
    def append(self, val):
        "Standard list append method"
        self[len(self):] = [val]

    def extend(self, vals):
        "Standard list extend method"
        self[len(self):] = vals

    def insert(self, index, val):
        "Standard list insert method"
        if not isinstance(index, int):
            raise TypeError("%s is not a legal index" % index)
        self[index:index] = [val]

    def pop(self, index=-1):
        "Standard list pop method"
        result = self[index]
        del self[index]
        return result

    def remove(self, val):
        "Standard list remove method"
        del self[self.index(val)]

    def reverse(self):
        "Standard list reverse method"
        self[:] = self[-1::-1]

    def sort(self, key=None, reverse=False):
        "Standard list sort method"
        self[:] = sorted(self, key=key, reverse=reverse)

    # ### Private routines ###
    def _rebuild(self, newLen, newItems):
        if newLen and newLen < self._minlength:
            raise ValueError('Must have at least %d items' % self._minlength)
        if self._maxlength is not None and newLen > self._maxlength:
            raise ValueError('Cannot have more than %d items' % self._maxlength)

        self._set_list(newLen, newItems)

    def _set_single_rebuild(self, index, value):
        self._set_slice(slice(index, index + 1, 1), [value])

    def _checkindex(self, index):
        length = len(self)
        if 0 <= index < length:
            return index
        if -length <= index < 0:
            return index + length
        raise IndexError('invalid index: %s' % index)

    def _check_allowed(self, items):
        if hasattr(self, '_allowed'):
            if False in [isinstance(val, self._allowed) for val in items]:
                raise TypeError('Invalid type encountered in the arguments.')

    def _set_slice(self, index, values):
        "Assign values to a slice of the object"
        try:
            valueList = list(values)
        except TypeError:
            raise TypeError('can only assign an iterable to a slice')

        self._check_allowed(valueList)

        origLen = len(self)
        start, stop, step = index.indices(origLen)

        # CAREFUL: index.step and step are not the same!
        # step will never be None
        if index.step is None:
            self._assign_simple_slice(start, stop, valueList)
        else:
            self._assign_extended_slice(start, stop, step, valueList)

    def _assign_extended_slice_rebuild(self, start, stop, step, valueList):
        'Assign an extended slice by rebuilding entire list'
        indexList = range(start, stop, step)
        # extended slice, only allow assigning slice of same size
        if len(valueList) != len(indexList):
            raise ValueError('attempt to assign sequence of size %d '
                             'to extended slice of size %d'
                             % (len(valueList), len(indexList)))

        # we're not changing the length of the sequence
        newLen = len(self)
        newVals = dict(zip(indexList, valueList))

        def newItems():
            for i in range(newLen):
                if i in newVals:
                    yield newVals[i]
                else:
                    yield self._get_single_internal(i)

        self._rebuild(newLen, newItems())

    def _assign_extended_slice(self, start, stop, step, valueList):
        'Assign an extended slice by re-assigning individual items'
        indexList = range(start, stop, step)
        # extended slice, only allow assigning slice of same size
        if len(valueList) != len(indexList):
            raise ValueError('attempt to assign sequence of size %d '
                             'to extended slice of size %d'
                             % (len(valueList), len(indexList)))

        for i, val in zip(indexList, valueList):
            self._set_single(i, val)

    def _assign_simple_slice(self, start, stop, valueList):
        'Assign a simple slice; Can assign slice of any length'
        origLen = len(self)
        stop = max(start, stop)
        newLen = origLen - stop + start + len(valueList)

        def newItems():
            for i in range(origLen + 1):
                if i == start:
                    yield from valueList

                if i < origLen:
                    if i < start or i >= stop:
                        yield self._get_single_internal(i)

        self._rebuild(newLen, newItems())



import datetime
import random

from faker.utils.checksums import calculate_luhn

from .. import Provider as SsnProvider


class Provider(SsnProvider):

    @staticmethod
    def _org_to_vat(org_id):
        org_id = org_id.replace('-', '')
        if len(org_id) == 10:
            org_id = '16' + org_id
        return 'SE{}01'.format(org_id)

    def ssn(self, min_age=18, max_age=90, long=False, dash=True):
        """
        Returns a 10 or 12 (long=True) digit Swedish SSN, "Personnummer".

        It consists of 10 digits in the form (CC)YYMMDD-SSSQ, where
        YYMMDD is the date of birth, SSS is a serial number
        and Q is a control character (Luhn checksum).

        Specifying dash=False will give a purely numeric string, suitable
        for writing direct to databases.

        http://en.wikipedia.org/wiki/Personal_identity_number_(Sweden)
        """

        age = datetime.timedelta(
            days=self.generator.random.randrange(min_age * 365, max_age * 365))
        birthday = datetime.datetime.now() - age
        yr_fmt = '%Y' if long else '%y'
        pnr_date = birthday.strftime('{}%m%d'.format(yr_fmt))
        chk_date = pnr_date[2:] if long else pnr_date
        suffix = str(self.generator.random.randrange(0, 999)).zfill(3)
        luhn_checksum = str(calculate_luhn(chk_date + suffix))
        hyphen = '-' if dash else ''
        pnr = '{}{}{}{}'.format(pnr_date, hyphen, suffix, luhn_checksum)

        return pnr

    ORG_ID_DIGIT_1 = (1, 2, 3, 5, 6, 7, 8, 9)

    def org_id(self, long=False, dash=True):
        """
        Returns a 10 or 12 digit Organisation ID for a Swedish
        company.
        (In Swedish) https://sv.wikipedia.org/wiki/Organisationsnummer
        """
        first_digits = list(self.ORG_ID_DIGIT_1)
        random.shuffle(first_digits)
        onr_one = str(first_digits.pop())
        onr_one += str(self.generator.random.randrange(0, 9)).zfill(1)
        onr_one += str(self.generator.random.randrange(20, 99))
        onr_one += str(self.generator.random.randrange(0, 99)).zfill(2)
        onr_two = str(self.generator.random.randrange(0, 999)).zfill(3)
        luhn_checksum = str(calculate_luhn(onr_one + onr_two))
        prefix = '16' if long else ''
        hyphen = '-' if dash else ''

        org_id = '{}{}{}{}{}'.format(prefix, onr_one, hyphen, onr_two, luhn_checksum)
        return org_id

    def vat_id(self):
        """
        http://ec.europa.eu/taxation_customs/vies/faq.html#item_11
        :return: A random Swedish VAT ID, based on a valid Org ID
        """
        oid = self.org_id(long=True, dash=False)
        vid = Provider._org_to_vat(oid)
        return vid

    def org_and_vat_id(self, long=False, dash=True):
        """Returns matching Org ID and VAT number"""
        oid = self.org_id(long=long, dash=dash)
        vid = Provider._org_to_vat(oid)
        return oid, vid



from . import router


class CacheConfig:
    ALL = "all"
    return_all = False

    def __init__(self, values, instance_values):
        self.values = values or []
        self.instance_values = instance_values or []


class CacheGroup:
    """
    Tracks versions for a collection of models.

    Each instance has a primary version and as many
    secondary versions as you need.

    Secondary versions are automatically cleared when the
    primary version is incremented.

    :param key: This should be unique as this will be used \
    to prefix all version values for this group in your cache.

    :param cache_name: The name of the cache you would like to \
    use. Defaults to DEFAULT_CACHE_ALIAS.

    :param version_expiry: How long before the version expires. \
    default is 30 days. This needs to be at least twice \
    as long as the longest lasting cache entry.
    """

    def __init__(self, key=None, version_expiry=None):
        assert key
        self.key = key

        self.route_group = "default"

        # Version expiry needs to be at least twice
        # as long as the longest lasting cache entry
        self.version_expiry = version_expiry
        if not self.version_expiry:
            self.version_expiry = 60 * 60 * 24 * 30
        self._models = {}

    def _get_cache(self, key):
        return router.router.get_cache(route_group=self.route_group, key=key)

    def register_models(self, *models, **kwargs):
        """
        Register multiple models with the same
        arguments.

        Calls register for each argument passed along with
        all keyword arguments.
        """

        for model in models:
            self.register(model, **kwargs)

    def register(self, model, values=None, instance_values=None):
        """
        Registers a model with this group.

        :param values: A list of values that should be incremented \
        whenever invalidate_cache is called for a instance or class \
        of this type.

        :param instance_values: A list of attribute names that will \
        be looked up on the instance of this model that is passed to \
        invalidate_cache. The value resulting from that lookup \
        will then be incremented.
        """

        if model in self._models:
            raise Exception(f"{model} is already registered")

        self._models[model] = CacheConfig(values, instance_values)

    def _get_models(self):
        return list(self._models.keys())

    models = property(_get_models)

    def _get_cache_extras(self, klass, instance=None, extra=None, force_all=False):
        # Return the extra keys for this klass instance combo
        # Return self.ALL to use the main version.
        if force_all:
            return CacheConfig.ALL

        config = self._models.get(klass)
        result = None

        if config:
            if not config.instance_values and not config.values:
                result = CacheConfig.ALL
            else:
                result = list(config.values)
                for value in config.instance_values:
                    if hasattr(instance, value):
                        result.append(getattr(instance, value))

                if extra:
                    result.extend(extra)

        return result

    def invalidate_cache(self, klass, instance=None, extra=None, force_all=False):
        """
        Use this method to invalidate keys related to a particular
        model or instance. Invalidating a cache is really just
        incrementing the version for the right key(s).

        :param klass: The model class you are invalidating. If the given \
        class was not registered with this group no action will be taken.

        :param instance: The instance you want to use with the registered\
        instance_values. Usually the instance that was just saved. \
        Defaults to None.

        :param extra: A list of extra values that you would like incremented \
        in addition to what was registered for this model.

        :param force_all: Ignore all registered values and provided \
        arguments and increment the major version for this group.
        """

        values = self._get_cache_extras(
            klass, instance=instance, extra=extra, force_all=force_all
        )

        if values == CacheConfig.ALL:
            self._increment_version()
        elif values:
            for value in values:
                self._increment_version(extra=value)

    def _get_extra_key(self, extra):
        # An extra key is based on the main version
        # plus the extra value. So that if the main
        # version changes the keys do too.
        v = self._get_cache(self.key).get(self.key)
        if v is None:
            # Set the base key, otherwise extras
            # that are created first won't be flushed
            # with all.
            v = self._increment_version()

        return f"{self.key}.{v}.{extra}"

    def get_version(self, extra=None):
        """
        This will return a string that can be used as a prefix
        for django's cache key. Something like key.1 or key.1.2

        If a version was not found '1' will be stored and returned as
        the number for that key.

        If extra is given a version will be returned for that value.
        Otherwise the major version will be returned.

        :param extra: the minor version to get. Defaults to None.
        """

        if extra:
            key = self._get_extra_key(extra)
        else:
            key = self.key

        v = self._get_cache(key).get(key)
        if v is None:
            v = self._increment_version(extra=extra)

        return f"{key}.{v}"

    def _increment_version(self, extra=None):
        # If we made this redis only we could switch to
        # using hashes to save memory. For now this is fine.
        if extra:
            key = self._get_extra_key(extra)
        else:
            key = self.key

        cache = self._get_cache(key)
        try:
            val = cache.incr(key)
            if val > 1000 * 1000:
                val = 0
                cache.set(key, val, timeout=self.version_expiry)
        except ValueError:
            val = 0
            cache.set(key, val, timeout=self.version_expiry)
        return val



"""Generate and work with PEP 425 Compatibility Tags."""

import distutils.util
import platform
import sys
import os
import sysconfig
import warnings

from .macosx_libfile import extract_macosx_min_system_version


try:
    from importlib.machinery import all_suffixes as get_all_suffixes
except ImportError:
    from imp import get_suffixes

    def get_all_suffixes():
        return [suffix[0] for suffix in get_suffixes()]


def get_config_var(var):
    try:
        return sysconfig.get_config_var(var)
    except IOError as e:  # pip Issue #1074
        warnings.warn("{0}".format(e), RuntimeWarning)
        return None


def get_abbr_impl():
    """Return abbreviated implementation name."""
    impl = platform.python_implementation()
    if impl == 'PyPy':
        return 'pp'
    elif impl == 'Jython':
        return 'jy'
    elif impl == 'IronPython':
        return 'ip'
    elif impl == 'CPython':
        return 'cp'

    raise LookupError('Unknown Python implementation: ' + impl)


def get_impl_ver():
    """Return implementation version."""
    impl_ver = get_config_var("py_version_nodot")
    if not impl_ver:
        impl_ver = ''.join(map(str, get_impl_version_info()))
    return impl_ver


def get_impl_version_info():
    """Return sys.version_info-like tuple for use in decrementing the minor
    version."""
    return sys.version_info[0], sys.version_info[1]


def get_flag(var, fallback, expected=True, warn=True):
    """Use a fallback method for determining SOABI flags if the needed config
    var is unset or unavailable."""
    val = get_config_var(var)
    if val is None:
        if warn:
            warnings.warn("Config variable '{0}' is unset, Python ABI tag may "
                          "be incorrect".format(var), RuntimeWarning, 2)
        return fallback()
    return val == expected


def get_abi_tag():
    """Return the ABI tag based on SOABI (if available) or emulate SOABI
    (CPython 2, PyPy)."""
    soabi = get_config_var('SOABI')
    impl = get_abbr_impl()
    if not soabi and impl in ('cp', 'pp') and hasattr(sys, 'maxunicode'):
        d = ''
        m = ''
        u = ''
        if get_flag('Py_DEBUG',
                    lambda: hasattr(sys, 'gettotalrefcount'),
                    warn=(impl == 'cp')):
            d = 'd'
        if get_flag('WITH_PYMALLOC',
                    lambda: impl == 'cp',
                    warn=(impl == 'cp' and
                          sys.version_info < (3, 8))) \
                and sys.version_info < (3, 8):
            m = 'm'
        if get_flag('Py_UNICODE_SIZE',
                    lambda: sys.maxunicode == 0x10ffff,
                    expected=4,
                    warn=(impl == 'cp' and
                          sys.version_info < (3, 3))) \
                and sys.version_info < (3, 3):
            u = 'u'
        abi = '%s%s%s%s%s' % (impl, get_impl_ver(), d, m, u)
    elif soabi and soabi.startswith('cpython-'):
        abi = 'cp' + soabi.split('-')[1]
    elif soabi:
        abi = soabi.replace('.', '_').replace('-', '_')
    else:
        abi = None
    return abi


def calculate_macosx_platform_tag(archive_root, platform_tag):
    """
    Calculate proper macosx platform tag basing on files which are included to wheel

    Example platform tag `macosx-10.14-x86_64`
    """
    prefix, base_version, suffix = platform_tag.split('-')
    base_version = tuple([int(x) for x in base_version.split(".")])
    if len(base_version) >= 2:
        base_version = base_version[0:2]

    assert len(base_version) == 2
    if "MACOSX_DEPLOYMENT_TARGET" in os.environ:
        deploy_target = tuple([int(x) for x in os.environ[
            "MACOSX_DEPLOYMENT_TARGET"].split(".")])
        if len(deploy_target) >= 2:
            deploy_target = deploy_target[0:2]
        if deploy_target < base_version:
            sys.stderr.write(
                 "[WARNING] MACOSX_DEPLOYMENT_TARGET is set to a lower value ({}) than the "
                 "version on which the Python interpreter was compiled ({}), and will be "
                 "ignored.\n".format('.'.join(str(x) for x in deploy_target),
                                     '.'.join(str(x) for x in base_version))
                )
        else:
            base_version = deploy_target

    assert len(base_version) == 2
    start_version = base_version
    versions_dict = {}
    for (dirpath, dirnames, filenames) in os.walk(archive_root):
        for filename in filenames:
            if filename.endswith('.dylib') or filename.endswith('.so'):
                lib_path = os.path.join(dirpath, filename)
                min_ver = extract_macosx_min_system_version(lib_path)
                if min_ver is not None:
                    versions_dict[lib_path] = min_ver[0:2]

    if len(versions_dict) > 0:
        base_version = max(base_version, max(versions_dict.values()))

    # macosx platform tag do not support minor bugfix release
    fin_base_version = "_".join([str(x) for x in base_version])
    if start_version < base_version:
        problematic_files = [k for k, v in versions_dict.items() if v > start_version]
        problematic_files = "\n".join(problematic_files)
        if len(problematic_files) == 1:
            files_form = "this file"
        else:
            files_form = "these files"
        error_message = \
            "[WARNING] This wheel needs a higher macOS version than {}  " \
            "To silence this warning, set MACOSX_DEPLOYMENT_TARGET to at least " +\
            fin_base_version + " or recreate " + files_form + " with lower " \
            "MACOSX_DEPLOYMENT_TARGET:  \n" + problematic_files

        if "MACOSX_DEPLOYMENT_TARGET" in os.environ:
            error_message = error_message.format("is set in MACOSX_DEPLOYMENT_TARGET variable.")
        else:
            error_message = error_message.format(
                "the version your Python interpreter is compiled against.")

        sys.stderr.write(error_message)

    platform_tag = prefix + "_" + fin_base_version + "_" + suffix
    return platform_tag


def get_platform(archive_root):
    """Return our platform name 'win32', 'linux_x86_64'"""
    # XXX remove distutils dependency
    result = distutils.util.get_platform()
    if result.startswith("macosx") and archive_root is not None:
        result = calculate_macosx_platform_tag(archive_root, result)
    result = result.replace('.', '_').replace('-', '_')
    if result == "linux_x86_64" and sys.maxsize == 2147483647:
        # pip pull request #3497
        result = "linux_i686"

    return result


def get_supported(archive_root, versions=None, supplied_platform=None):
    """Return a list of supported tags for each version specified in
    `versions`.

    :param versions: a list of string versions, of the form ["33", "32"],
        or None. The first version will be assumed to support our ABI.
    """
    supported = []

    # Versions must be given with respect to the preference
    if versions is None:
        versions = []
        version_info = get_impl_version_info()
        major = version_info[:-1]
        # Support all previous minor Python versions.
        for minor in range(version_info[-1], -1, -1):
            versions.append(''.join(map(str, major + (minor,))))

    impl = get_abbr_impl()

    abis = []

    abi = get_abi_tag()
    if abi:
        abis[0:0] = [abi]

    abi3s = set()
    for suffix in get_all_suffixes():
        if suffix.startswith('.abi'):
            abi3s.add(suffix.split('.', 2)[1])

    abis.extend(sorted(list(abi3s)))

    abis.append('none')

    platforms = []
    if supplied_platform:
        platforms.append(supplied_platform)
    platforms.append(get_platform(archive_root))

    # Current version, current API (built specifically for our Python):
    for abi in abis:
        for arch in platforms:
            supported.append(('%s%s' % (impl, versions[0]), abi, arch))

    # abi3 modules compatible with older version of Python
    for version in versions[1:]:
        # abi3 was introduced in Python 3.2
        if version in ('31', '30'):
            break
        for abi in abi3s:   # empty set if not Python 3
            for arch in platforms:
                supported.append(("%s%s" % (impl, version), abi, arch))

    # No abi / arch, but requires our implementation:
    for i, version in enumerate(versions):
        supported.append(('%s%s' % (impl, version), 'none', 'any'))
        if i == 0:
            # Tagged specifically as being cross-version compatible
            # (with just the major version specified)
            supported.append(('%s%s' % (impl, versions[0][0]), 'none', 'any'))

    # Major Python version + platform; e.g. binaries not using the Python API
    for arch in platforms:
        supported.append(('py%s' % (versions[0][0]), 'none', arch))

    # No abi / arch, generic Python
    for i, version in enumerate(versions):
        supported.append(('py%s' % (version,), 'none', 'any'))
        if i == 0:
            supported.append(('py%s' % (version[0]), 'none', 'any'))

    return supported



from django.core.exceptions import ObjectDoesNotExist
from django.db.models import signals
from django.db.models.aggregates import *  # NOQA
from django.db.models.aggregates import __all__ as aggregates_all
from django.db.models.constraints import *  # NOQA
from django.db.models.constraints import __all__ as constraints_all
from django.db.models.deletion import (
    CASCADE, DO_NOTHING, PROTECT, SET, SET_DEFAULT, SET_NULL, ProtectedError,
)
from django.db.models.expressions import (
    Case, Exists, Expression, ExpressionList, ExpressionWrapper, F, Func,
    OuterRef, RowRange, Subquery, Value, ValueRange, When, Window, WindowFrame,
)
from django.db.models.fields import *  # NOQA
from django.db.models.fields import __all__ as fields_all
from django.db.models.fields.files import FileField, ImageField
from django.db.models.fields.proxy import OrderWrt
from django.db.models.indexes import *  # NOQA
from django.db.models.indexes import __all__ as indexes_all
from django.db.models.lookups import Lookup, Transform
from django.db.models.manager import Manager
from django.db.models.query import (
    Prefetch, Q, QuerySet, prefetch_related_objects,
)
from django.db.models.query_utils import FilteredRelation

# Imports that would create circular imports if sorted
from django.db.models.base import DEFERRED, Model  # isort:skip
from django.db.models.fields.related import (  # isort:skip
    ForeignKey, ForeignObject, OneToOneField, ManyToManyField,
    ManyToOneRel, ManyToManyRel, OneToOneRel,
)


__all__ = aggregates_all + constraints_all + fields_all + indexes_all
__all__ += [
    'ObjectDoesNotExist', 'signals',
    'CASCADE', 'DO_NOTHING', 'PROTECT', 'SET', 'SET_DEFAULT', 'SET_NULL',
    'ProtectedError',
    'Case', 'Exists', 'Expression', 'ExpressionList', 'ExpressionWrapper', 'F',
    'Func', 'OuterRef', 'RowRange', 'Subquery', 'Value', 'ValueRange', 'When',
    'Window', 'WindowFrame',
    'FileField', 'ImageField', 'OrderWrt', 'Lookup', 'Transform', 'Manager',
    'Prefetch', 'Q', 'QuerySet', 'prefetch_related_objects', 'DEFERRED', 'Model',
    'FilteredRelation',
    'ForeignKey', 'ForeignObject', 'OneToOneField', 'ManyToManyField',
    'ManyToOneRel', 'ManyToManyRel', 'OneToOneRel',
]



from django import forms
from django.contrib.admin.utils import label_for_field
from django.core.exceptions import ObjectDoesNotExist
from django.db import models
from django.db.models.fields import FieldDoesNotExist
from django.forms.utils import ErrorList, flatatt
from django.utils.encoding import force_text
from django.utils.html import conditional_escape
from django.utils.safestring import mark_safe
from django.utils.text import capfirst

from . import fields, widgets

FORMFIELD_FOR_DBFIELD_DEFAULTS = {
    models.ForeignKey: {"widget": widgets.APIChoiceWidget},
    models.ManyToManyField: {"widget": widgets.APIManyChoiceWidget},
    models.DateTimeField: {"widget": widgets.DateTimeWidget},
    models.DateField: {"widget": widgets.DateWidget},
}


class AdminList:
    ASC = "asc"
    DESC = "desc"

    def __init__(
        self,
        formset,
        object_list,
        visible_fields,
        sort_fields,
        order_types,
        model_name=None,
    ):
        self.formset = formset
        self.object_list = object_list
        self.visible_fields = visible_fields
        self.order_types = order_types
        self.sort_fields = sort_fields
        self.model_name = model_name
        self.empty = len(object_list) == 0
        self.link_num = 0
        if formset:
            for i, field in enumerate(visible_fields):
                if not formset.empty_form.fields.get(field):
                    self.link_num = i
                    break

        self.auto_sort = False
        if self.formset:
            self.auto_sort = (
                len(
                    [
                        x
                        for x in list(formset.empty_form.fields.values())
                        if isinstance(x, fields.OrderFormField)
                    ]
                )
                > 0
            )

    def __iter__(self):
        if self.formset:
            for form in self.formset:
                yield AdminListRow(form, form.instance, self.visible_fields)
        else:
            for obj in self.object_list:
                yield AdminListRow(None, obj, self.visible_fields)

    def labels(self):
        """
        Get field label for fields
        """

        if isinstance(self.object_list, list):
            model = self.formset.model
        else:
            model = self.object_list.model

        for field in self.visible_fields:
            name = None
            if self.formset:
                f = self.formset.empty_form.fields.get(field, None)
                if f:
                    name = f.label

            if name is None:
                name = label_for_field(field, model)

            if name == model._meta.verbose_name:
                name = self.model_name and self.model_name or model._meta.verbose_name

            stype = None
            cur_sorted = False

            sortable = False

            if self.order_types:
                sortable = get_sort_field(field, model)
                stype = self.ASC

                # change order_type so that next sorting on the same
                # field will give reversed results
                if sortable and field in self.sort_fields:
                    field_index = self.sort_fields.index(field)
                    cur_sorted = True
                    if self.order_types[field_index] == self.ASC:
                        stype = self.DESC
                    elif self.order_types[field_index] == self.DESC:
                        stype = self.ASC
                    else:
                        stype = self.ASC

            yield AdminListLabel(name, field, stype, cur_sorted, bool(sortable))


class AdminListLabel:
    def __init__(self, name, attr, order_type, cur_sorted, sortable):
        self.name = name
        self.attr = attr
        self.order_type = order_type
        self.cur_sorted = cur_sorted
        self.sortable = sortable


class AdminListRow:
    def __init__(self, form, instance, visible_fields):
        self.form = form
        self.instance = instance
        self.visible_fields = visible_fields

    def get_value(self, field, i):
        if self.form and self.form.fields.get(field):
            return AdminField(self.form, field, is_first=(i == 0))
        else:
            return get_field_value(field, self.instance)

    def __iter__(self):
        for i, field in enumerate(self.visible_fields):
            yield self.get_value(field, i)


class AdminForm:
    def __init__(self, form, fieldsets):
        self.form, self.fieldsets = form, normalize_fieldsets(fieldsets)

    def __iter__(self):
        for name, options in self.fieldsets:
            yield Fieldset(self.form, name, **options)

    def first_field(self):
        try:
            fieldset_name, fieldset_options = self.fieldsets[0]
            field_name = fieldset_options["fields"][0]
            if not isinstance(field_name, str):
                field_name = field_name[0]
            return self.form[field_name]
        except (KeyError, IndexError):
            pass
        try:
            return next(iter(self.form))
        except StopIteration:
            return None


class AdminFormSets:
    def __init__(self, formsets, combined_defs):
        self.formsets = formsets
        self.combined_sets = []
        self.hidden_sets = set()

        if combined_defs:
            if isinstance(combined_defs, dict):
                self._add_combined(combined_defs)

            elif hasattr(combined_defs, "__iter__"):
                for set_def in combined_defs:
                    if not isinstance(set_def, dict):
                        continue

                    self._add_combined(set_def)

    def _add_combined(self, set_def):
        keys = set_def.get("keys")
        self.hidden_sets = self.hidden_sets.union(set(keys))
        self.combined_sets.append(CombinedMultiFormSet(self, **set_def))

    def visible_formsets(self):
        for v in self.combined_sets:
            yield v.title, v
        for k, v in list(self.formsets.items()):
            if k not in self.hidden_sets:
                yield k, v

    def all_formsets(self):
        for k, v in list(self.formsets.items()):
            yield k, v


class CombinedMultiFormSet:
    def __init__(self, admin_formset, keys=None, order_by=None, title=None):
        if not order_by:
            order_by = "order"

        self.title = title
        self.keys = keys
        self.order_by = order_by
        self.admin_formset = admin_formset
        self.combined = True

    @property
    def formsets(self):
        return self.admin_formset.formsets

    def management_form(self):
        return mark_safe(
            "".join(
                [
                    str(x.management_form)
                    for k, x in list(self.formsets.items())
                    if k in self.keys
                ]
            )
        )

    def non_form_errors(self):
        return mark_safe(
            "".join(
                [
                    str(x.non_form_errors())
                    for k, x in list(self.formsets.items())
                    if k in self.keys
                ]
            )
        )

    def __iter__(self):
        forms = []
        for k, formset in list(self.formsets.items()):
            forms = forms + [(x, k) for x in formset]

        forms = sorted(forms, key=lambda f: getattr(f[0].instance, self.order_by))
        for f in forms:
            yield f[0], self.formsets[f[1]]


class Fieldset:
    def __init__(self, form, name=None, fields=(), classes=(), description=None):
        self.form = form
        self.name, self.fields = name, fields
        self.classes = " ".join(classes)
        self.description = description

    def __iter__(self):
        for field in self.fields:
            yield Fieldline(self.form, field)


class Fieldline:
    def __init__(self, form, field):
        self.form = form  # A django.forms.Form instance
        if isinstance(field, tuple) or isinstance(field, list):
            self.fields = [item for item in field]
        else:
            self.fields = [field]

    def __iter__(self):
        for i, field in enumerate(self.fields):
            cls = ReadOnlyField
            if isinstance(field, str) and field in self.form.fields:
                cls = AdminField
            yield cls(self.form, field, is_first=(i == 0))

    def errors(self):
        fields = [f for f in self.fields if f in self.form.fields]
        return mark_safe("\n".join([self.form[f].errors.as_ul() for f in fields]))


class AdminField:
    def __init__(self, form, field, is_first):
        self.field = form[field]  # A django.forms.BoundField instance
        self.is_first = is_first  # Whether this field is first on the line
        self.is_checkbox = isinstance(self.field.field.widget, forms.CheckboxInput)
        self.is_date = isinstance(self.field.field, forms.DateField) or isinstance(
            self.field.field, forms.SplitDateTimeField
        )

        self.is_order_field = isinstance(self.field.field, fields.OrderFormField)

    def label_tag(self):
        classes = []
        contents = conditional_escape(force_text(self.field.label))
        if self.is_checkbox:
            classes.append("vCheckboxLabel")
        else:
            contents += ":"
        if self.field.field.required:
            classes.append("required")
        if not self.is_first:
            classes.append("inline")
        attrs = classes and {"class": " ".join(classes)} or {}
        return self.field.label_tag(contents=mark_safe(contents), attrs=attrs)

    def errors(self):
        return mark_safe(self.field.errors.as_ul())


class InnerField:
    def __init__(self, field, instance):
        if callable(field):
            class_name = field.__name__ != "<lambda>" and field.__name__ or ""
        else:
            class_name = field

        self.name = class_name
        self.label = label_for_field(field, instance.__class__)
        self.field_repr = get_field_value(field, instance)
        self.help_text = get_field_attr(field, instance, "help_text", "")

    def __str__(self):
        return force_text(self.field_repr)


class ReadOnlyField:
    def __init__(self, form, field, is_first):
        self.field = InnerField(field, form.instance)
        self.form = form
        self.is_first = is_first
        self.is_checkbox = False
        self.is_readonly = True

    def label_tag(self):
        attrs = {}
        if not self.is_first:
            attrs["class"] = "inline"
        try:
            label = self.field.label
            return mark_safe(
                "<label{0}>{1}:</label>".format(
                    conditional_escape(flatatt(attrs)),
                    capfirst(conditional_escape(force_text(label))),
                )
            )
        except Exception:
            return ""

    def errors(self):
        return ErrorList()


def normalize_fieldsets(fieldsets):
    """
    Make sure the keys in fieldset dictionaries are strings. Returns the
    normalized data.
    """
    result = []
    for name, options in fieldsets:
        result.append((name, normalize_dictionary(options)))
    return result


def normalize_dictionary(data_dict):
    """
    Converts all the keys in "data_dict" to strings. The keys must be
    convertible using str().
    """
    for key, value in list(data_dict.items()):
        if not isinstance(key, str):
            del data_dict[key]
            data_dict[str(key)] = value
    return data_dict


def get_sort_field(attr, model):
    """
    Get's the field to sort on for the given
    attr.

    Currently returns attr if it is a field on
    the given model.

    If the models has an attribute matching that name
    and that value has an attribute 'sort_field' than
    that value is used.

    TODO: Provide a way to sort based on a non field
    attribute.
    """

    try:
        if model._meta.get_field(attr):
            return attr
    except FieldDoesNotExist:
        if isinstance(attr, str):
            val = getattr(model, attr, None)
            if val and hasattr(val, "sort_field"):
                return getattr(model, attr).sort_field
        return None


def check_unicode(instance, field):
    if hasattr(instance, "__str__"):
        return instance.__str__()
    elif hasattr(instance, "__unicode__"):
        # Python 2 compatible
        return instance.__unicode__()
    else:
        return ""


def get_field_value(field, instance):
    if field == "__unicode__":
        return check_unicode(instance, field)
    try:
        f = instance._meta.get_field(field)
        try:
            value = getattr(instance, field)
        except (AttributeError, ObjectDoesNotExist):
            value = None

        if hasattr(f, "flatchoices") and f.flatchoices:
            value = dict(f.flatchoices).get(value)
        elif isinstance(value, models.Manager):
            value = ", ".join([str(x) for x in value.all()])

    except models.FieldDoesNotExist:
        # For non-field values, the value is either a method, property or
        # returned via a callable.
        if callable(field):
            value = field(instance)
        else:
            attr = getattr(instance, field)
            if callable(attr):
                value = attr()
            else:
                value = attr
    return value


def get_field_attr(field, instance, attr, default=""):
    value = default
    try:
        f = instance._meta.get_field(field)
        value = getattr(f, attr, default)
    except models.FieldDoesNotExist:
        if callable(field):
            value = getattr(field, attr, default)
        else:
            p = getattr(instance, field)
            if callable(p):
                value = getattr(p, attr, default)
    return value


def pluralize(value, custom_plural):
    if custom_plural:
        return custom_plural
    elif value.lower().endswith("s"):
        return value + "es"
    else:
        return value + "s"


def model_name(
    model, custom_model_name=None, custom_model_name_plural=None, plural=False
):

    if custom_model_name:
        value = custom_model_name
        if plural:
            value = pluralize(custom_model_name, custom_model_name_plural)
    else:
        value = model._meta.verbose_name
        if plural:
            value = model._meta.verbose_name_plural
    return capfirst_if_needed(value)


def capfirst_if_needed(value):
    if value and not value[0].isupper():
        value = capfirst(value)
    return value



from django.contrib import messages
from django.contrib.auth import authenticate, login, REDIRECT_FIELD_NAME
from django.contrib.auth.forms import PasswordChangeForm
from django.contrib.auth.views import logout as Signout
from django.http import Http404, HttpResponseForbidden
from django.shortcuts import get_object_or_404, redirect
from django.urls import reverse
from django.utils.translation import ugettext as _
from django.views.generic import FormView, TemplateView, View
from django.views.generic.list import ListView, MultipleObjectMixin

from . import settings as accounts_settings
from . import signals as accounts_signals
from .decorators import secure_required
from .forms import (
    AuthenticationForm,
    ChangeEmailForm,
    EditProfileForm,
    SignupForm
)
from .models import AccountsSignup
from .utils import get_profile_model, get_user_model, signin_redirect


class ExtraContextTemplateView(TemplateView):
    """ Add extra context to a simple template view """

    extra_context = None

    def get_context_data(self, *args, **kwargs):
        context = super().get_context_data(
            *args, **kwargs
        )
        if self.extra_context:
            context.update(self.extra_context)
        return context

    # this view is used in POST requests,
    # e.g. signup when the form is not valid
    post = TemplateView.get


@secure_required
def activate(
    request,
    activation_key,
    template_name="accounts/activate_fail.html",
    success_url=None,
    extra_context=None,
):
    """
    Activate a user with an activation key.

    The key is a SHA1 string. When the SHA1 is found with an
    :class:`AccountsSignup`, the :class:`User` of that account will be
    activated.  After a successful activation the view will redirect to
    ``success_url``.  If the SHA1 is not found, the user will be shown the
    ``template_name`` template displaying a fail message.

    :param activation_key:
        String of a SHA1 string of 40 characters long. A SHA1 is always 160bit
        long, with 4 bits per character this makes it --160/4-- 40 characters
        long.

    :param template_name:
        String containing the template name that is used when the
        ``activation_key`` is invalid and the activation fails. Defaults to
        ``accounts/activation_fail.html``.

    :param success_url:
        String containing the URL where the user should be redirected to after
        a successful activation. Will replace ``%(username)s`` with string
        formatting if supplied. If ``success_url`` is left empty, will direct
        to ``accounts_profile_detail`` view.

    :param extra_context:
        Dictionary containing variables which could be added to the template
        context. Default to an empty dictionary.

    """
    user = AccountsSignup.objects.activate_user(activation_key)
    if user:
        # Sign the user in.
        auth_user = authenticate(identification=user.email, check_password=False)
        login(request, auth_user)

        if accounts_settings.ACCOUNTS_USE_MESSAGES:
            messages.success(
                request,
                _("Your account has been activated and you have been signed in."),
                fail_silently=True,
            )

        if success_url:
            redirect_to = success_url % {"username": user.username}
        else:
            redirect_to = reverse(
                "accounts_profile_detail", kwargs={"username": user.username}
            )
        return redirect(redirect_to)
    else:
        if not extra_context:
            extra_context = dict()
        return ExtraContextTemplateView.as_view(
            template_name=template_name, extra_context=extra_context
        )(request)


@secure_required
def email_confirm(
    request,
    confirmation_key,
    template_name="accounts/email_confirm_fail.html",
    success_url=None,
    extra_context=None,
):
    """
    Confirms an email address with a confirmation key.

    Confirms a new email address by running :func:`User.objects.confirm_email`
    method. If the method returns an :class:`User` the user will have his new
    e-mail address set and redirected to ``success_url``. If no ``User`` is
    returned the user will be represented with a fail message from
    ``template_name``.

    :param confirmation_key:
        String with a SHA1 representing the confirmation key used to verify a
        new email address.

    :param template_name:
        String containing the template name which should be rendered when
        confirmation fails. When confirmation is successful, no template is
        needed because the user will be redirected to ``success_url``.

    :param success_url:
        String containing the URL which is redirected to after a successful
        confirmation.  Supplied argument must be able to be rendered by
        ``reverse`` function.

    :param extra_context:
        Dictionary of variables that are passed on to the template supplied by
        ``template_name``.

    """
    user = AccountsSignup.objects.confirm_email(confirmation_key)
    if user:
        if accounts_settings.ACCOUNTS_USE_MESSAGES:
            messages.success(
                request, _("Your email address has been changed."), fail_silently=True
            )

        if success_url:
            redirect_to = success_url
        else:
            redirect_to = reverse(
                "accounts_email_confirm_complete", kwargs={"username": user.username}
            )
        return redirect(redirect_to)
    else:
        if not extra_context:
            extra_context = dict()
        return ExtraContextTemplateView.as_view(
            template_name=template_name, extra_context=extra_context
        )(request)


def direct_to_user_template(request, username, template_name, extra_context=None):
    """
    Simple wrapper for Django's :func:`direct_to_template` view.

    This view is used when you want to show a template to a specific user. A
    wrapper for :func:`direct_to_template` where the template also has access
    to the user that is found with ``username``. For ex. used after signup,
    activation and confirmation of a new e-mail.

    :param username:
        String defining the username of the user that made the action.

    :param template_name:
        String defining the name of the template to use. Defaults to
        ``accounts/signup_complete.html``.

    **Keyword arguments**

    ``extra_context``
        A dictionary containing extra variables that should be passed to the
        rendered template. The ``account`` key is always the ``User``
        that completed the action.

    **Extra context**

    ``viewed_user``
        The currently :class:`User` that is viewed.

    """
    user = get_object_or_404(get_user_model(), username__iexact=username)

    if not extra_context:
        extra_context = dict()
    extra_context["viewed_user"] = user
    extra_context["profile"] = user.get_profile()
    return ExtraContextTemplateView.as_view(
        template_name=template_name, extra_context=extra_context
    )(request)


@secure_required
def signin(
    request,
    auth_form=AuthenticationForm,
    template_name="accounts/signin_form.html",
    redirect_field_name=REDIRECT_FIELD_NAME,
    redirect_signin_function=signin_redirect,
    extra_context=None,
):
    """
    Signin using email or username with password.

    Signs a user in by combining email/username with password. If the
    combination is correct and the user :func:`is_active` the
    :func:`redirect_signin_function` is called with the arguments
    ``REDIRECT_FIELD_NAME`` and an instance of the :class:`User` who is is
    trying the login. The returned value of the function will be the URL that
    is redirected to.

    A user can also select to be remembered for ``ACCOUNTS_REMEMBER_DAYS``.

    :param auth_form:
        Form to use for signing the user in. Defaults to the
        :class:`AuthenticationForm` supplied by accounts.

    :param template_name:
        String defining the name of the template to use. Defaults to
        ``accounts/signin_form.html``.

    :param redirect_field_name:
        Form field name which contains the value for a redirect to the
        succeeding page. Defaults to ``next`` and is set in
        ``REDIRECT_FIELD_NAME`` setting.

    :param redirect_signin_function:
        Function which handles the redirect. This functions gets the value of
        ``REDIRECT_FIELD_NAME`` and the :class:`User` who has logged in. It
        must return a string which specifies the URI to redirect to.

    :param extra_context:
        A dictionary containing extra variables that should be passed to the
        rendered template. The ``form`` key is always the ``auth_form``.

    **Context**

    ``form``
        Form used for authentication supplied by ``auth_form``.

    """
    form = auth_form()

    if request.method == "POST":
        form = auth_form(request.POST, request.FILES)
        if form.is_valid():
            identification = form.cleaned_data["identification"]
            password = form.cleaned_data["password"]
            remember_me = form.cleaned_data["remember_me"]

            user = authenticate(identification=identification, password=password)
            if user.is_active:
                login(request, user)
                if remember_me:
                    request.session.set_expiry(
                        accounts_settings.ACCOUNTS_REMEMBER_ME_DAYS[1] * 86400
                    )
                else:
                    request.session.set_expiry(0)

                if accounts_settings.ACCOUNTS_USE_MESSAGES:
                    messages.success(
                        request, _("You have been signed in."), fail_silently=True
                    )

                # Whereto now?
                redirect_to = redirect_signin_function(
                    request.GET.get(redirect_field_name), user
                )
                return redirect(redirect_to)
            else:
                return redirect(
                    reverse("accounts_disabled", kwargs={"username": user.username})
                )

    if not extra_context:
        extra_context = dict()
    extra_context.update(
        {"form": form, "next": request.GET.get(redirect_field_name),}
    )
    return ExtraContextTemplateView.as_view(
        template_name=template_name, extra_context=extra_context
    )(request)


@secure_required
def signout(
    request,
    next_page=accounts_settings.ACCOUNTS_REDIRECT_ON_SIGNOUT,
    template_name="accounts/signout.html",
    *args,
    **kwargs
):
    """
    Signs out the user and adds a success message ``You have been signed
    out.`` If next_page is defined you will be redirected to the URI. If
    not the template in template_name is used.

    :param next_page:
        A string which specifies the URI to redirect to.

    :param template_name:
        String defining the name of the template to use. Defaults to
        ``accounts/signout.html``.

    """
    if (
        request.user.is_authenticated() and accounts_settings.ACCOUNTS_USE_MESSAGES
    ):  # pragma: no cover
        messages.success(request, _("You have been signed out."), fail_silently=True)
    return Signout(request, next_page, template_name, *args, **kwargs)


@secure_required
def email_change(
    request,
    username,
    email_form=ChangeEmailForm,
    template_name="accounts/email_form.html",
    success_url=None,
    extra_context=None,
):
    """
    Change email address

    :param username:
        String of the username which specifies the current account.

    :param email_form:
        Form that will be used to change the email address. Defaults to
        :class:`ChangeEmailForm` supplied by accounts.

    :param template_name:
        String containing the template to be used to display the email form.
        Defaults to ``accounts/email_form.html``.

    :param success_url:
        Named URL where the user will get redirected to when successfully
        changing their email address.  When not supplied will redirect to
        ``accounts_email_complete`` URL.

    :param extra_context:
        Dictionary containing extra variables that can be used to render the
        template. The ``form`` key is always the form supplied by the keyword
        argument ``form`` and the ``user`` key by the user whose email address
        is being changed.

    **Context**

    ``form``
        Form that is used to change the email address supplied by ``form``.

    ``account``
        Instance of the ``Account`` whose email address is about to be changed.

    **Todo**

    Need to have per-object permissions, which enables users with the correct
    permissions to alter the email address of others.

    """
    user = get_object_or_404(get_user_model(), username__iexact=username)

    form = email_form(user)

    if request.method == "POST":
        form = email_form(user, request.POST, request.FILES)

        if form.is_valid():
            form.save()

            if success_url:
                redirect_to = success_url
            else:
                redirect_to = reverse(
                    "accounts_email_change_complete", kwargs={"username": user.username}
                )
            return redirect(redirect_to)

    if not extra_context:
        extra_context = dict()
    extra_context["form"] = form
    extra_context["profile"] = user.get_profile()
    return ExtraContextTemplateView.as_view(
        template_name=template_name, extra_context=extra_context
    )(request)


@secure_required
def password_change(
    request,
    username,
    template_name="accounts/password_form.html",
    pass_form=PasswordChangeForm,
    success_url=None,
    extra_context=None,
):
    """ Change password of user.

    This view is almost a mirror of the view supplied in
    :func:`contrib.auth.views.password_change`, with the minor change that in
    this view we also use the username to change the password. This was needed
    to keep our URLs logical (and REST) across the entire application. And
    that in a later stadium administrators can also change the users password
    through the web application itself.

    :param username:
        String supplying the username of the user who's password is about to be
        changed.

    :param template_name:
        String of the name of the template that is used to display the password
        change form. Defaults to ``accounts/password_form.html``.

    :param pass_form:
        Form used to change password. Default is the form supplied by Django
        itself named ``PasswordChangeForm``.

    :param success_url:
        Named URL that is passed onto a :func:`reverse` function with
        ``username`` of the active user. Defaults to the
        ``accounts_password_complete`` URL.

    :param extra_context:
        Dictionary of extra variables that are passed on to the template. The
        ``form`` key is always used by the form supplied by ``pass_form``.

    **Context**

    ``form``
        Form used to change the password.

    """
    user = get_object_or_404(get_user_model(), username__iexact=username)

    form = pass_form(user=user)

    if request.method == "POST":
        form = pass_form(user=user, data=request.POST)
        if form.is_valid():
            form.save()

            # Send a signal that the password has changed
            accounts_signals.password_complete.send(sender=None, user=user)

            if success_url:
                redirect_to = success_url
            else:
                redirect_to = reverse(
                    "accounts_password_change_complete",
                    kwargs={"username": user.username},
                )
            return redirect(redirect_to)

    if not extra_context:
        extra_context = dict()
    extra_context["form"] = form
    extra_context["profile"] = user.get_profile()
    return ExtraContextTemplateView.as_view(
        template_name=template_name, extra_context=extra_context
    )(request)


@secure_required
def profile_edit(
    request,
    username,
    edit_profile_form=EditProfileForm,
    template_name="accounts/profile_form.html",
    success_url=None,
    extra_context=None,
    **kwargs
):
    """
    Edit profile.

    Edits a profile selected by the supplied username. First checks
    permissions if the user is allowed to edit this profile, if denied will
    show a 404. When the profile is successfully edited will redirect to
    ``success_url``.

    :param username:
        Username of the user which profile should be edited.

    :param edit_profile_form:

        Form that is used to edit the profile. The :func:`EditProfileForm.save`
        method of this form will be called when the form
        :func:`EditProfileForm.is_valid`.  Defaults to :class:`EditProfileForm`
        from accounts.

    :param template_name:
        String of the template that is used to render this view. Defaults to
        ``accounts/edit_profile_form.html``.

    :param success_url:
        Named URL which will be passed on to a django ``reverse`` function
        after the form is successfully saved. Defaults to the
        ``accounts_detail`` url.

    :param extra_context:
        Dictionary containing variables that are passed on to the
        ``template_name`` template.  ``form`` key will always be the form used
        to edit the profile, and the ``profile`` key is always the edited
        profile.

    **Context**

    ``form``
        Form that is used to alter the profile.

    ``profile``
        Instance of the ``Profile`` that is edited.

    """
    user = get_object_or_404(get_user_model(), username__iexact=username)

    profile = user.get_profile()

    user_initial = {"first_name": user.first_name, "last_name": user.last_name}

    form = edit_profile_form(instance=profile, initial=user_initial)

    if request.method == "POST":
        form = edit_profile_form(
            request.POST, request.FILES, instance=profile, initial=user_initial
        )

        if form.is_valid():
            profile = form.save()

            if accounts_settings.ACCOUNTS_USE_MESSAGES:
                messages.success(
                    request, _("Your profile has been updated."), fail_silently=True
                )

            if success_url:
                redirect_to = success_url
            else:
                redirect_to = reverse(
                    "accounts_profile_detail", kwargs={"username": username}
                )
            return redirect(redirect_to)

    if not extra_context:
        extra_context = dict()
    extra_context["form"] = form
    extra_context["profile"] = profile
    return ExtraContextTemplateView.as_view(
        template_name=template_name, extra_context=extra_context
    )(request)


def profile_detail(
    request,
    username,
    template_name=accounts_settings.ACCOUNTS_PROFILE_DETAIL_TEMPLATE,
    extra_context=None,
    **kwargs
):
    """
    Detailed view of an user.

    :param username:
        String of the username of which the profile should be viewed.

    :param template_name:
        String representing the template name that should be used to display
        the profile.

    :param extra_context:
        Dictionary of variables which should be supplied to the template. The
        ``profile`` key is always the current profile.

    **Context**

    ``profile``
        Instance of the currently viewed ``Profile``.

    """
    user = get_object_or_404(get_user_model(), username__iexact=username)
    profile_model = get_profile_model()
    try:
        profile = user.get_profile()
    except profile_model.DoesNotExist:
        profile = profile_model(user=user)
        profile.save()
    if not profile.can_view_profile(request.user):
        return HttpResponseForbidden(
            _("You don't have permission to view this profile.")
        )
    if not extra_context:
        extra_context = dict()
    extra_context["profile"] = user.get_profile()
    return ExtraContextTemplateView.as_view(
        template_name=template_name, extra_context=extra_context
    )(request)


def account_delete(
    request,
    username,
    template_name=accounts_settings.ACCOUNTS_PROFILE_DETAIL_TEMPLATE,
    extra_context=None,
    **kwargs
):
    """
    Delete an account.
    """
    user = get_object_or_404(get_user_model(), username__iexact=username)
    user.is_active = False
    user.save()

    return redirect(reverse("accounts_admin"))


class ProfileListView(ListView):
    """
    Lists all profiles
    """

    context_object_name = "profile_list"
    page = 1
    paginate_by = 20
    template_name = "accounts/profile_list.html"
    extra_context = None

    def get_context_data(self, **kwargs):
        # Call the base implementation first to get a context
        context = super().get_context_data(**kwargs)
        try:
            page = int(self.request.GET.get("page", None))
        except (TypeError, ValueError):
            page = self.page

        if not self.request.user.is_staff:
            raise Http404

        if not self.extra_context:
            self.extra_context = dict()

        context["page"] = page
        context["paginate_by"] = self.paginate_by
        context["extra_context"] = self.extra_context

        context["form"] = SignupForm()

        return context

    def get_queryset(self):
        profile_model = get_profile_model()
        queryset = profile_model.objects.get_visible_profiles(self.request.user)
        return queryset


class AccountsFormView(FormView, MultipleObjectMixin):
    template_name = "accounts/profile_list.html"
    form_class = SignupForm

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        return context

    def get_success_url(self):
        return reverse("accounts_admin", kwargs=None)

    def form_valid(self, form):
        if not self.request.user.is_authenticated():
            return HttpResponseForbidden()

        user = form.save()

        # Send the signup complete signal
        accounts_signals.signup_complete.send(sender=None, user=user)

        # record the interest using the message in form.cleaned_data
        return super().form_valid(form)


class AccountsListView(View):
    def get(self, request, *args, **kwargs):
        view = ProfileListView.as_view()
        return view(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        view = AccountsFormView.as_view()
        return view(request, *args, **kwargs)



import os
from datetime import datetime
from urllib.parse import urljoin

from django.conf import settings
from django.core.exceptions import SuspiciousFileOperation
from django.core.files import File, locks
from django.core.files.move import file_move_safe
from django.core.signals import setting_changed
from django.utils import timezone
from django.utils._os import safe_join
from django.utils.crypto import get_random_string
from django.utils.deconstruct import deconstructible
from django.utils.encoding import filepath_to_uri
from django.utils.functional import LazyObject, cached_property
from django.utils.module_loading import import_string
from django.utils.text import get_valid_filename

__all__ = (
    'Storage', 'FileSystemStorage', 'DefaultStorage', 'default_storage',
    'get_storage_class',
)


class Storage:
    """
    A base storage class, providing some default behaviors that all other
    storage systems can inherit or override, as necessary.
    """

    # The following methods represent a public interface to private methods.
    # These shouldn't be overridden by subclasses unless absolutely necessary.

    def open(self, name, mode='rb'):
        """Retrieve the specified file from storage."""
        return self._open(name, mode)

    def save(self, name, content, max_length=None):
        """
        Save new content to the file specified by name. The content should be
        a proper File object or any Python file-like object, ready to be read
        from the beginning.
        """
        # Get the proper name for the file, as it will actually be saved.
        if name is None:
            name = content.name

        if not hasattr(content, 'chunks'):
            content = File(content, name)

        name = self.get_available_name(name, max_length=max_length)
        return self._save(name, content)

    # These methods are part of the public API, with default implementations.

    def get_valid_name(self, name):
        """
        Return a filename, based on the provided filename, that's suitable for
        use in the target storage system.
        """
        return get_valid_filename(name)

    def get_available_name(self, name, max_length=None):
        """
        Return a filename that's free on the target storage system and
        available for new content to be written to.
        """
        dir_name, file_name = os.path.split(name)
        file_root, file_ext = os.path.splitext(file_name)
        # If the filename already exists, add an underscore and a random 7
        # character alphanumeric string (before the file extension, if one
        # exists) to the filename until the generated filename doesn't exist.
        # Truncate original name if required, so the new filename does not
        # exceed the max_length.
        while self.exists(name) or (max_length and len(name) > max_length):
            # file_ext includes the dot.
            name = os.path.join(dir_name, "%s_%s%s" % (file_root, get_random_string(7), file_ext))
            if max_length is None:
                continue
            # Truncate file_root if max_length exceeded.
            truncation = len(name) - max_length
            if truncation > 0:
                file_root = file_root[:-truncation]
                # Entire file_root was truncated in attempt to find an available filename.
                if not file_root:
                    raise SuspiciousFileOperation(
                        'Storage can not find an available filename for "%s". '
                        'Please make sure that the corresponding file field '
                        'allows sufficient "max_length".' % name
                    )
                name = os.path.join(dir_name, "%s_%s%s" % (file_root, get_random_string(7), file_ext))
        return name

    def generate_filename(self, filename):
        """
        Validate the filename by calling get_valid_name() and return a filename
        to be passed to the save() method.
        """
        # `filename` may include a path as returned by FileField.upload_to.
        dirname, filename = os.path.split(filename)
        return os.path.normpath(os.path.join(dirname, self.get_valid_name(filename)))

    def path(self, name):
        """
        Return a local filesystem path where the file can be retrieved using
        Python's built-in open() function. Storage systems that can't be
        accessed using open() should *not* implement this method.
        """
        raise NotImplementedError("This backend doesn't support absolute paths.")

    # The following methods form the public API for storage systems, but with
    # no default implementations. Subclasses must implement *all* of these.

    def delete(self, name):
        """
        Delete the specified file from the storage system.
        """
        raise NotImplementedError('subclasses of Storage must provide a delete() method')

    def exists(self, name):
        """
        Return True if a file referenced by the given name already exists in the
        storage system, or False if the name is available for a new file.
        """
        raise NotImplementedError('subclasses of Storage must provide an exists() method')

    def listdir(self, path):
        """
        List the contents of the specified path. Return a 2-tuple of lists:
        the first item being directories, the second item being files.
        """
        raise NotImplementedError('subclasses of Storage must provide a listdir() method')

    def size(self, name):
        """
        Return the total size, in bytes, of the file specified by name.
        """
        raise NotImplementedError('subclasses of Storage must provide a size() method')

    def url(self, name):
        """
        Return an absolute URL where the file's contents can be accessed
        directly by a Web browser.
        """
        raise NotImplementedError('subclasses of Storage must provide a url() method')

    def get_accessed_time(self, name):
        """
        Return the last accessed time (as a datetime) of the file specified by
        name. The datetime will be timezone-aware if USE_TZ=True.
        """
        raise NotImplementedError('subclasses of Storage must provide a get_accessed_time() method')

    def get_created_time(self, name):
        """
        Return the creation time (as a datetime) of the file specified by name.
        The datetime will be timezone-aware if USE_TZ=True.
        """
        raise NotImplementedError('subclasses of Storage must provide a get_created_time() method')

    def get_modified_time(self, name):
        """
        Return the last modified time (as a datetime) of the file specified by
        name. The datetime will be timezone-aware if USE_TZ=True.
        """
        raise NotImplementedError('subclasses of Storage must provide a get_modified_time() method')


@deconstructible
class FileSystemStorage(Storage):
    """
    Standard filesystem storage
    """
    # The combination of O_CREAT and O_EXCL makes os.open() raise OSError if
    # the file already exists before it's opened.
    OS_OPEN_FLAGS = os.O_WRONLY | os.O_CREAT | os.O_EXCL | getattr(os, 'O_BINARY', 0)

    def __init__(self, location=None, base_url=None, file_permissions_mode=None,
                 directory_permissions_mode=None):
        self._location = location
        self._base_url = base_url
        self._file_permissions_mode = file_permissions_mode
        self._directory_permissions_mode = directory_permissions_mode
        setting_changed.connect(self._clear_cached_properties)

    def _clear_cached_properties(self, setting, **kwargs):
        """Reset setting based property values."""
        if setting == 'MEDIA_ROOT':
            self.__dict__.pop('base_location', None)
            self.__dict__.pop('location', None)
        elif setting == 'MEDIA_URL':
            self.__dict__.pop('base_url', None)
        elif setting == 'FILE_UPLOAD_PERMISSIONS':
            self.__dict__.pop('file_permissions_mode', None)
        elif setting == 'FILE_UPLOAD_DIRECTORY_PERMISSIONS':
            self.__dict__.pop('directory_permissions_mode', None)

    def _value_or_setting(self, value, setting):
        return setting if value is None else value

    @cached_property
    def base_location(self):
        return self._value_or_setting(self._location, settings.MEDIA_ROOT)

    @cached_property
    def location(self):
        return os.path.abspath(self.base_location)

    @cached_property
    def base_url(self):
        if self._base_url is not None and not self._base_url.endswith('/'):
            self._base_url += '/'
        return self._value_or_setting(self._base_url, settings.MEDIA_URL)

    @cached_property
    def file_permissions_mode(self):
        return self._value_or_setting(self._file_permissions_mode, settings.FILE_UPLOAD_PERMISSIONS)

    @cached_property
    def directory_permissions_mode(self):
        return self._value_or_setting(self._directory_permissions_mode, settings.FILE_UPLOAD_DIRECTORY_PERMISSIONS)

    def _open(self, name, mode='rb'):
        return File(open(self.path(name), mode))

    def _save(self, name, content):
        full_path = self.path(name)

        # Create any intermediate directories that do not exist.
        directory = os.path.dirname(full_path)
        if not os.path.exists(directory):
            try:
                if self.directory_permissions_mode is not None:
                    # Set the umask because os.makedirs() doesn't apply the "mode"
                    # argument to intermediate-level directories.
                    old_umask = os.umask(0o777 & ~self.directory_permissions_mode)
                    try:
                        os.makedirs(directory, self.directory_permissions_mode)
                    finally:
                        os.umask(old_umask)
                else:
                    os.makedirs(directory)
            except FileExistsError:
                # There's a race between os.path.exists() and os.makedirs().
                # If os.makedirs() fails with FileExistsError, the directory
                # was created concurrently.
                pass
        if not os.path.isdir(directory):
            raise IOError("%s exists and is not a directory." % directory)

        # There's a potential race condition between get_available_name and
        # saving the file; it's possible that two threads might return the
        # same name, at which point all sorts of fun happens. So we need to
        # try to create the file, but if it already exists we have to go back
        # to get_available_name() and try again.

        while True:
            try:
                # This file has a file path that we can move.
                if hasattr(content, 'temporary_file_path'):
                    file_move_safe(content.temporary_file_path(), full_path)

                # This is a normal uploadedfile that we can stream.
                else:
                    # The current umask value is masked out by os.open!
                    fd = os.open(full_path, self.OS_OPEN_FLAGS, 0o666)
                    _file = None
                    try:
                        locks.lock(fd, locks.LOCK_EX)
                        for chunk in content.chunks():
                            if _file is None:
                                mode = 'wb' if isinstance(chunk, bytes) else 'wt'
                                _file = os.fdopen(fd, mode)
                            _file.write(chunk)
                    finally:
                        locks.unlock(fd)
                        if _file is not None:
                            _file.close()
                        else:
                            os.close(fd)
            except FileExistsError:
                # A new name is needed if the file exists.
                name = self.get_available_name(name)
                full_path = self.path(name)
            else:
                # OK, the file save worked. Break out of the loop.
                break

        if self.file_permissions_mode is not None:
            os.chmod(full_path, self.file_permissions_mode)

        # Store filenames with forward slashes, even on Windows.
        return name.replace('\\', '/')

    def delete(self, name):
        assert name, "The name argument is not allowed to be empty."
        name = self.path(name)
        # If the file or directory exists, delete it from the filesystem.
        try:
            if os.path.isdir(name):
                os.rmdir(name)
            else:
                os.remove(name)
        except FileNotFoundError:
            # FileNotFoundError is raised if the file or directory was removed
            # concurrently.
            pass

    def exists(self, name):
        return os.path.exists(self.path(name))

    def listdir(self, path):
        path = self.path(path)
        directories, files = [], []
        for entry in os.scandir(path):
            if entry.is_dir():
                directories.append(entry.name)
            else:
                files.append(entry.name)
        return directories, files

    def path(self, name):
        return safe_join(self.location, name)

    def size(self, name):
        return os.path.getsize(self.path(name))

    def url(self, name):
        if self.base_url is None:
            raise ValueError("This file is not accessible via a URL.")
        url = filepath_to_uri(name)
        if url is not None:
            url = url.lstrip('/')
        return urljoin(self.base_url, url)

    def _datetime_from_timestamp(self, ts):
        """
        If timezone support is enabled, make an aware datetime object in UTC;
        otherwise make a naive one in the local timezone.
        """
        if settings.USE_TZ:
            # Safe to use .replace() because UTC doesn't have DST
            return datetime.utcfromtimestamp(ts).replace(tzinfo=timezone.utc)
        else:
            return datetime.fromtimestamp(ts)

    def get_accessed_time(self, name):
        return self._datetime_from_timestamp(os.path.getatime(self.path(name)))

    def get_created_time(self, name):
        return self._datetime_from_timestamp(os.path.getctime(self.path(name)))

    def get_modified_time(self, name):
        return self._datetime_from_timestamp(os.path.getmtime(self.path(name)))


def get_storage_class(import_path=None):
    return import_string(import_path or settings.DEFAULT_FILE_STORAGE)


class DefaultStorage(LazyObject):
    def _setup(self):
        self._wrapped = get_storage_class()()


default_storage = DefaultStorage()



import datetime
import urllib.error
import urllib.parse
import urllib.request

from django import forms
from django.contrib.admin.widgets import url_params_from_lookup_dict
from django.forms import widgets
from django.forms.utils import flatatt
from django.template.loader import render_to_string
from django.urls import reverse
from django.utils import formats, timezone, translation
from django.utils.dateparse import parse_time
from django.utils.encoding import force_text, python_2_unicode_compatible
from django.utils.html import conditional_escape, format_html, html_safe
from django.utils.safestring import mark_safe

from . import settings

# NOT EVERYTHING IS SUPPORTED, I DON'T CARE.
TRANSLATION_DICT = {
    # Day
    "d": "dd",
    "l": "DD",
    "j": "oo",
    # Month
    "B": "MM",
    "m": "mm",
    "b": "M",
    # Year
    "Y": "yy",
    "y": "y",
    # Time
    "p": "TT",
    "I": "hh",
    "H": "HH",
    "M": "mm",
    "S": "ss",
}


@html_safe
@python_2_unicode_compatible
class SubWidget:
    """
    Some widgets are made of multiple HTML elements -- namely, RadioSelect.
    This is a class that represents the "inner" HTML element of a widget.
    """

    def __init__(self, parent_widget, name, value, attrs, choices):
        self.parent_widget = parent_widget
        self.name, self.value = name, value
        self.attrs, self.choices = attrs, choices

    def __str__(self):
        args = [self.name, self.value, self.attrs]
        if self.choices:
            args.append(self.choices)
        return self.parent_widget.render(*args)


@html_safe
@python_2_unicode_compatible
class ChoiceInput(SubWidget):
    """
    An object used by ChoiceFieldRenderer that represents a single
    <input type='$input_type'>.
    """

    input_type = None  # Subclasses must define this

    def __init__(self, name, value, attrs, choice, index):
        self.name = name
        self.value = value
        self.attrs = attrs
        self.choice_value = force_text(choice[0])
        self.choice_label = force_text(choice[1])
        self.index = index
        if "id" in self.attrs:
            self.attrs["id"] += f"_{self.index}"

    def __str__(self):
        return self.render()

    def render(self, name=None, value=None, attrs=None, renderer=None):
        if self.id_for_label:
            label_for = format_html(' for="{}"', self.id_for_label)
        else:
            label_for = ""
        attrs = dict(self.attrs, **attrs) if attrs else self.attrs
        return format_html(
            "<label{}>{} {}</label>", label_for, self.tag(attrs), self.choice_label
        )

    def is_checked(self):
        return self.value == self.choice_value

    def tag(self, attrs=None):
        attrs = attrs or self.attrs
        final_attrs = dict(
            attrs, type=self.input_type, name=self.name, value=self.choice_value
        )
        if self.is_checked():
            final_attrs["checked"] = "checked"
        return format_html("<input{} />", flatatt(final_attrs))

    @property
    def id_for_label(self):
        return self.attrs.get("id", "")


class RadioChoiceInput(ChoiceInput):
    input_type = "radio"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.value = force_text(self.value)


@html_safe
@python_2_unicode_compatible
class ChoiceFieldRenderer:
    """
    An object used by RadioSelect to enable customization of radio widgets.
    """

    choice_input_class = None
    outer_html = "<ul{id_attr}>{content}</ul>"
    inner_html = "<li>{choice_value}{sub_widgets}</li>"

    def __init__(self, name, value, attrs, choices):
        self.name = name
        self.value = value
        self.attrs = attrs
        self.choices = choices

    def __getitem__(self, idx):
        return list(self)[idx]

    def __iter__(self):
        for idx, choice in enumerate(self.choices):
            yield self.choice_input_class(
                self.name, self.value, self.attrs.copy(), choice, idx
            )

    def __str__(self):
        return self.render()

    def render(self):
        """
        Outputs a <ul> for this set of choice fields.
        If an id was given to the field, it is applied to the <ul> (each
        item in the list will get an id of `$id_$i`).
        """
        id_ = self.attrs.get("id")
        output = []
        for i, choice in enumerate(self.choices):
            choice_value, choice_label = choice
            if isinstance(choice_label, (tuple, list)):
                attrs_plus = self.attrs.copy()
                if id_:
                    attrs_plus["id"] += "_{}".format(i)
                sub_ul_renderer = self.__class__(
                    name=self.name,
                    value=self.value,
                    attrs=attrs_plus,
                    choices=choice_label,
                )
                sub_ul_renderer.choice_input_class = self.choice_input_class
                output.append(
                    format_html(
                        self.inner_html,
                        choice_value=choice_value,
                        sub_widgets=sub_ul_renderer.render(),
                    )
                )
            else:
                w = self.choice_input_class(
                    self.name, self.value, self.attrs.copy(), choice, i
                )
                output.append(
                    format_html(
                        self.inner_html, choice_value=force_text(w), sub_widgets=""
                    )
                )
        return format_html(
            self.outer_html,
            id_attr=format_html(' id="{}"', id_) if id_ else "",
            content=mark_safe("\n".join(output)),
        )


class RadioFieldRenderer(ChoiceFieldRenderer):
    choice_input_class = RadioChoiceInput


def translate_format(format_string):
    for k, v in list(TRANSLATION_DICT.items()):
        format_string = format_string.replace("%{0}".format(k), v)
    return format_string


class DateWidget(widgets.DateInput):
    bc = "date"
    format_key = "DATE_INPUT_FORMATS"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.attrs["class"] = self.bc
        if "format" not in kwargs:
            self.format = settings.DATE_INPUT_FORMATS

    def get_format(self):
        if self.format:
            return self.format

        if settings.USE_SCARLET_DATE_FORMATS and hasattr(settings, self.format_key):
            return getattr(settings, self.format_key)

        return formats.get_format(self.format_key)[0]

    def _format_value(self, value):
        return formats.localize_input(value, self.get_format())

    def build_attrs(self, *args, **kwargs):
        args = super().build_attrs(*args, **kwargs)
        args["data-date-format"] = translate_format(self.get_format())
        args["data-timezone"] = timezone.get_current_timezone_name()
        args["data-locale"] = translation.get_language()
        args["autocomplete"] = 'off'
        return args

    def value_from_datadict(self, data, files, name):
        value = super().value_from_datadict(data, files, name)
        df = self.get_format()
        if isinstance(value, str):
            try:
                return datetime.datetime.strptime(value, df)
            except ValueError:
                pass
        return value


class DateTimeWidget(DateWidget):
    bc = "datetime"
    format_key = "DATETIME_INPUT_FORMATS"


class TimeChoiceWidget(widgets.Select):
    """
    Widget for time fields. A select widget that will have a 'now'
    option plus an option for each block of time you want to
    display. By default this will be one item in the drop down for
    every 15 min block of a day.

    :param attrs: HTML attributes for the widget; same as django's.
    :param min_interval: Interval for minutes in your dropdown, \
    should be a number between 1 and 60. Default is 15.
    :param sec_interval: Interval for seconds in your dropdown, \
    should be a number between 1 and 60. Default is 60.
    :param twenty_four_hour: Display time in a 24hr format? \
    Default is False.
    """

    NOW = "now"

    def __init__(
        self, attrs=None, min_interval=15, sec_interval=60, twenty_four_hour=False
    ):
        super().__init__(attrs)
        assert 60 >= min_interval > 0
        assert 60 >= sec_interval > 0

        self.twenty_four_hour = twenty_four_hour
        self.choices = [(self.NOW, "Now")]
        self.choice_values = set()

        self.repr_format = "%I:%M:%S %p"
        if twenty_four_hour:
            self.repr_format = "%H:%M:%S"

        for hour in range(24):
            for min_v in range(60 // min_interval):
                min_v = min_v * min_interval
                for sec in range(60 // sec_interval):
                    sec = sec * sec_interval
                    t = datetime.time(hour, min_v, sec)
                    self.choices.append(
                        (t.strftime("%H:%M:%S"), t.strftime(self.repr_format))
                    )
                    self.choice_values.add(t.strftime("%H:%M:%S"))

    def value_from_datadict(self, *args):
        data = super().value_from_datadict(*args)
        if data == self.NOW:
            # Time should be naive, conversion happens later
            data = datetime.datetime.now().strftime("%H:%M:%S")
        return data

    def render(self, name, value, attrs=None, renderer=None, choices=()):
        if value:
            if isinstance(value, str):
                try:
                    value = parse_time(value)
                except ValueError:
                    value = None

            if value and isinstance(value, datetime.time):
                value_str = value.strftime("%H:%M:%S")
                if value_str not in self.choice_values:
                    choices = list(choices)
                    choices.append((value_str, value.strftime(self.repr_format)))
        return super().render(
            name, value, attrs=attrs, renderer=None, choices=choices
        )


class SplitDateTime(widgets.SplitDateTimeWidget):
    """
    Widget for datetime fields. Uses DateWidget, TimeChoiceWidget.
    """

    def __init__(self, widgets=(DateWidget, TimeChoiceWidget), attrs=None):
        forms.MultiWidget.__init__(self, widgets, attrs)

    def format_output(self, rendered_widgets):
        return mark_safe(f"{rendered_widgets[0]} {rendered_widgets[1]}")

    def value_from_datadict(self, data, files, name):
        d = super().value_from_datadict(data, files, name)
        if not self.is_required and len(d) and not d[0]:
            return ["", ""]
        return d


class DateRadioInput(RadioChoiceInput):
    label_text = "At a specific date and time"

    def render(self, name=None, value=None, attrs=None, renderer=None, choices=()):
        attrs = attrs or self.attrs
        if "id" in self.attrs:
            label_for = f' for="{self.attrs["id"]}_{self.index}"'
        else:
            label_for = ""
        date_widget = attrs["date_widget"]
        return mark_safe(
            f"<label{label_for}>{self.tag()} {self.label_text}: {date_widget}</label>"
        )

    def tag(self):
        final_attrs = {
            "type": "radio",
            "name": self.name,
            "value": self.choice_value,
        }

        if self.is_checked():
            final_attrs["checked"] = "checked"
        return mark_safe(f"<input{flatatt(final_attrs)} />")


class DateRenderer(RadioFieldRenderer):
    def __init__(self, *args, **kwargs):
        self.date_widget = kwargs.pop("date_widget")
        super().__init__(*args, **kwargs)

    def return_choice(self, choice, idx):
        cls = RadioChoiceInput
        attrs = self.attrs.copy()
        if choice[0] == RadioDateTimeWidget.DATE:
            cls = DateRadioInput
            attrs["date_widget"] = self.date_widget

        return cls(self.name, self.value, attrs, choice, idx)

    def __iter__(self):
        for i, choice in enumerate(self.choices):
            yield self.return_choice(choice, i)

    def __getitem__(self, idx):
        choice = self.choices[idx]
        return self.return_choice(choice, idx)

    def render(self):
        text = "\n".join([force_text(w) for w in self])
        return mark_safe(
            '<fieldset class="datetime">\n{}\n</fieldset>'.format(text)
        )


class RadioDateTimeWidget(widgets.RadioSelect):
    NOW = "now"
    DATE = "date"

    def __init__(self, *args, **kwargs):
        self.date_class = kwargs.pop("date_class", DateTimeWidget)
        self.choices = [(self.NOW, "Now"), (self.DATE, "Date",)]
        kwargs["choices"] = self.choices
        super().__init__(*args, **kwargs)

    def get_radio_key(self, name):
        return "{0}_rdi".format(name)

    def get_renderer(self, date_widget, name, value, attrs=None):
        return DateRenderer(name, value, attrs, self.choices, date_widget=date_widget)

    def render(self, name, value, attrs=None, renderer=None):
        widget = self.date_class()
        date_widget = widget.render(name, value, attrs=attrs, renderer=renderer)
        return self.get_renderer(
            date_widget, self.get_radio_key(name), self.DATE, {}
        ).render()

    def value_from_datadict(self, data, files, name):
        radio_value = data.get(self.get_radio_key(name))
        if radio_value == self.NOW:
            return timezone.now()
        else:
            widget = self.date_class()
            return widget.value_from_datadict(data, files, name)


class APIChoiceWidget(widgets.Input):
    """
    Widget for selecting a related object. This is used
    as the default widget for ForeignKey fields. Outputs
    text input field that is wrapped in a <div> that contains
    3 data attributes.

    * data-api: The url that can be queried to get the options \
    for this field in json format.
    * data-add: The url that should be called in a popup to add \
    a new item. If not present adding is not supported.
    * data-title: The title of the field.

    In order for this widget to work it needs to know where those urls are and
    if the rendering user has the needed permissions. This is
    accomplished by having the code that prepares the form
    call the `update_links` method. See the method documentation for
    what parameters are needed.

    :param rel: The rel attribute of the foreign key field that \
    this widget is for.
    :param attrs: HTML attributes for this field, same as django's.
    :param using: The database to use. Defaults to None.
    :param view: The string name of the view that will be used for \
    getting the api url. Defaults to 'main'.
    :param api_url: The api url. This is only used if the automatic url \
    discovery fails.
    :param add_view: The string name of the view that will be used for \
    getting the add url. Defaults to 'add'.
    :param add_url: The url for adding a new item. This is only used \
    if the automatic url discovery fails.
    :param extra_query_kwargs: Keyword arguments that you would like \
    passed as part of the query string.
    """

    input_type = "hidden"
    template = '<div class="api-select" data-title="%(value)s" data-api="%(link)s" data-add="%(add_link)s">%(input)s</div>'

    def __init__(
        self,
        rel,
        attrs=None,
        using=None,
        view="main",
        api_url="",
        add_view="add",
        add_url="",
        extra_query_kwargs=None,
    ):
        super().__init__(attrs=attrs)
        # TODO: Follow new django convention and rename rel to remote_field.
        self.rel = rel
        self.model = self.rel.model
        self.db = using

        self.extra_query_kwargs = extra_query_kwargs

        self.view = view
        self.add_view = add_view

        self._api_link = api_url
        self._add_link = add_url

    def render(self, name, value, attrs=None, renderer=None, choices=()):
        data = {
            "input": super().render(name, value, attrs=attrs, renderer=renderer),
            "value": conditional_escape(self.label_for_value(value)),
            "link": self.get_api_link(),
            "add_link": self.get_add_link(),
        }
        return mark_safe(self.template % data)

    def get_qs(self):
        """
        Returns a mapping that will be used to generate
        the query string for the api url. Any values
        in the the `limit_choices_to` specified on the
        foreign key field and any arguments specified on
        self.extra_query_kwargs are converted to a format
        that can be used in a query string and returned as
        a dictionary.
        """
        qs = url_params_from_lookup_dict(self.rel.limit_choices_to)
        if not qs:
            qs = {}

        if self.extra_query_kwargs:
            qs.update(self.extra_query_kwargs)
        return qs

    def _get_bundle_link(self, bundle, view_name, user):
        url = bundle.get_view_url(view_name, user)
        if url:
            return url
        return ""

    def _get_reverse(self, name, url_kwargs):
        return reverse(name, kwargs=url_kwargs)

    def update_links(self, request, admin_site=None):
        """
        Called to update the widget's urls. Tries to find the
        bundle for the model that this foreign key points to and then
        asks it for the urls for adding and listing and sets them on
        this widget instance. The urls are only set if request.user
        has permissions on that url.

        :param request: The request for which this widget is being rendered.
        :param admin_site: If provided, the `admin_site` is used to lookup \
        the bundle that is registered as the primary url for the model \
        that this foreign key points to.
        """
        if admin_site:
            bundle = admin_site.get_bundle_for_model(self.model)

            if bundle:
                self._api_link = self._get_bundle_link(bundle, self.view, request.user)
                self._add_link = self._get_bundle_link(
                    bundle, self.add_view, request.user
                )

    def get_api_link(self):
        """
        Adds a query string to the api url. At minimum adds the type=choices
        argument so that the return format is json. Any other filtering
        arguments calculated by the `get_qs` method are then added to the
        url. It is up to the destination url to respect them as filters.
        """
        url = self._api_link
        if url:
            qs = self.get_qs()
            url = f"{url}?type=choices"
            if qs:
                url = f"{url}&amp;" + "&amp;".join(
                        [
                            f"{k}={urllib.parse.quote(str(v).encode('utf8'))}"
                            for k, v in list(qs.items())
                        ]
                    )

                url = f"{url}&amp;" + "&amp;".join([f"exclude={x}" for x in list(qs.keys())])
        return url

    def get_add_link(self):
        """
        Appends the popup=1 query string to the url so the
        destination url treats it as a popup.
        """
        url = self._add_link
        if url:
            return f"{url}?popup=1"
        return url

    def label_for_value(self, value, key=None):
        """
        Looks up the current value of the field and returns
        a unicode representation. Default implementation does a lookup
        on the target model and if a match is found calls force_text
        on that object. Otherwise a blank string is returned.
        """
        if not key:
            key = self.rel.get_related_field().name

        if value is not None:
            try:
                obj = self.model._default_manager.using(self.db).get(**{key: value})
                return force_text(obj)
            except (ValueError, self.model.DoesNotExist):
                return ""
        return ""


class APIModelChoiceWidget(APIChoiceWidget):
    """
    Widget for selecting a related object. This is meant to
    be used in forms that specify their own related fields.
    Inherits from APIChoiceWidget but is based on a model
    instead of a foreign key relation.

    :param model: The model that this widget is for.
    :param attrs: HTML attributes for this field, same as django's.
    :param using: The database to use. Defaults to None.
    :param limit_choices_to: Keyword arguments that you would like \
    passed as part of the query string.
    :param view: The string name of the view that will be used for \
    getting the api url. Defaults to 'main'.
    :param api_url: The api url. This is only used if the automatic url \
    discovery fails.
    :param add_view: The string name of the view that will be used for \
    getting the add url. Defaults to 'add'.
    :param add_url: The url for adding a new item. This is only used \
    if the automatic url discovery fails.
    """

    template = '<div class="api-select" data-title="%(value)s" data-api="%(link)s" data-add="%(add_link)s">%(input)s</div>'

    def __init__(
        self,
        model,
        attrs=None,
        using=None,
        limit_choices_to=None,
        view="main",
        api_url="",
        add_view="add",
        add_url="",
    ):
        super(APIChoiceWidget, self).__init__(attrs=attrs)
        self.limit_choices_to = limit_choices_to
        self.model = model
        self.db = using

        self.view = view
        self.add_view = add_view

        self._api_link = api_url
        self._add_link = add_url

    def get_qs(self):
        return url_params_from_lookup_dict(self.limit_choices_to)

    def label_for_value(self, value):
        return super().label_for_value(value, key="pk")


class APIManyChoiceWidget(APIChoiceWidget, widgets.SelectMultiple):
    """
    Widget for selecting a many related objects. This is meant to
    be used in forms that specify their own related fields.
    Inherits from APIChoiceWidget but is based on a model
    instead of a foreign key relation.

    :param model: The model that this widget is for.
    :param attrs: HTML attributes for this field, same as django's.
    :param using: The database to use. Defaults to None.
    :param limit_choices_to: Keyword arguments that you would like \
    passed as part of the query string.
    :param view: The string name of the view that will be used for \
    getting the api url. Defaults to 'main'.
    :param api_url: The api url. This is only used if the automatic url \
    discovery fails.
    :param add_view: The string name of the view that will be used for \
    getting the add url. Defaults to 'add'.
    :param add_url: The url for adding a new item. This is only used \
    if the automatic url discovery fails.
    """

    template = '<div class="api-select" data-api="%(api_link)s" data-add="%(add_link)s">%(options)s</div>'
    allow_multiple_selected = True

    def __init__(
        self,
        model,
        attrs=None,
        using=None,
        limit_choices_to={},
        view="main",
        api_url="",
        add_view="add",
        add_url="",
    ):
        super(APIChoiceWidget, self).__init__(attrs=attrs)
        self.limit_choices_to = limit_choices_to
        self.model = model
        self.db = using

        self.view = view
        self.add_view = add_view

        self._api_link = api_url
        self._add_link = add_url

    def get_qs(self):
        return url_params_from_lookup_dict(self.limit_choices_to)

    def update_links(self, request, admin_site=None):
        """
        Called to update the widget's urls. Tries to find the
        bundle for the model that this foreign key points to and then
        asks it for the urls for adding and listing and sets them on
        this widget instance. The urls are only set if request.user
        has permissions on that url.

        :param request: The request for which this widget is being rendered.
        :param admin_site: If provided, the `admin_site` is used to lookup \
        the bundle that is registered as the primary url for the model \
        that this foreign key points to.
        """
        if admin_site:
            bundle = admin_site.get_bundle_for_model(self.model.model)

            if bundle:
                self._api_link = self._get_bundle_link(bundle, self.view, request.user)
                self._add_link = self._get_bundle_link(
                    bundle, self.add_view, request.user
                )

    def render(self, name, value, attrs=None, renderer=None, choices=()):
        final_attrs = self.build_attrs(attrs, {name: name})
        data = {
            "api_link": self.get_api_link(),
            "add_link": self.get_add_link(),
            "options": self.get_options(value, name),
        }
        data.update(final_attrs)
        return mark_safe(self.template % data)

    def get_options(self, value, name, key=None):
        if not key:
            key = self.model.get_related_field().name

        values = []
        if value is not None:
            try:
                kwargs = {"{0}__in".format(key): value}
                if self.limit_choices_to:
                    kwargs.update(self.limit_choices_to)
                objs = self.model.model._default_manager.using(self.db).filter(**kwargs)
                for obj in objs:
                    d = {
                        "text": force_text(obj),
                        "value": getattr(obj, key),
                        "name": name,
                    }
                    line = (
                        '<input type="hidden" data-multiple data-title="%(text)s" name="%(name)s" value="%(value)s" />'
                        % d
                    )
                    values.append(line)
            except Exception:
                pass

        if not values:
            values = [
                '<input type="hidden" data-multiple data-title="" name="{0}" value="" />'.format(
                    name
                )
            ]
        return "".join(values)


class HiddenTextInput(widgets.HiddenInput):
    """
    Widget for order fields in lists. Inherits from HiddenInput
    so it is marked as hidden in the form, but uses a 'text' input
    type with a class attribute in the rendered html of
    *orderfield*.
    """

    input_type = "text"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.attrs["class"] = "orderfield"

    def is_hidden(self):
        return True


class HTMLWidget(widgets.Textarea):
    """
    WYSIWYG Widget. Adds *widget-wysiwyg* to the class attribute
    in the rendered html.
    """

    template = "cms/toolbar.html"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        classes = ["editor__textarea"]
        if self.attrs.get('class'):
            classes.append(self.attrs.get('class'))
        self.attrs['class'] = " ".join(classes)

    def render(self, *args, **kwargs):
        text = super().render(*args, **kwargs)
        return mark_safe(u"<div class=\"editor\">{1} <div class=\"editor__content\">{0}</div></div>".format(text, render_to_string(self.template)))


class AnnotatedHTMLWidget(widgets.MultiWidget):
    """
    Combines WYSIWYG with a hidden widget for seperating
    annotation data from annotated text.
    """

    template = "cms/toolbar_annotation.html"

    START_HTML = '<div class="editor__annotation-data">'
    END_HTML = '</div>'

    def __init__(self, attrs=None):
        _widgets = (
            widgets.Textarea(attrs={'class': "editor__textarea"}),
            widgets.Textarea(attrs={'class': "editor__annotations"}),
        )
        super().__init__(_widgets, attrs=attrs)

    def decompress(self, value):
        if value:
            parts = value.rpartition(self.START_HTML)
            if parts[1]:
                annotation = parts[2]
                if annotation.endswith(self.END_HTML):
                    annotation = annotation[: -len(self.END_HTML)]
                return parts[0], annotation
            return [value, ""]
        return ["", ""]

    def format_output(self, rendered_widgets):
        return mark_safe(u"<div class=\"editor editor--annotations\">{0} {1} {2}</div>".format(
            render_to_string(self.template), *rendered_widgets))

    def value_from_datadict(self, data, files, name):
        data = [
            widget.value_from_datadict(data, files, f"{name}_{i}")
            for i, widget in enumerate(self.widgets)
        ]
        if data and data[1]:
            data[1] = self.START_HTML + data[1] + self.END_HTML
        return data[0] + data[1]



from django.db.models.lookups import (
    Exact, GreaterThan, GreaterThanOrEqual, In, IsNull, LessThan,
    LessThanOrEqual,
)


class MultiColSource:
    contains_aggregate = False

    def __init__(self, alias, targets, sources, field):
        self.targets, self.sources, self.field, self.alias = targets, sources, field, alias
        self.output_field = self.field

    def __repr__(self):
        return "{}({}, {})".format(
            self.__class__.__name__, self.alias, self.field)

    def relabeled_clone(self, relabels):
        return self.__class__(relabels.get(self.alias, self.alias),
                              self.targets, self.sources, self.field)

    def get_lookup(self, lookup):
        return self.output_field.get_lookup(lookup)


def get_normalized_value(value, lhs):
    from django.db.models import Model
    if isinstance(value, Model):
        value_list = []
        sources = lhs.output_field.get_path_info()[-1].target_fields
        for source in sources:
            while not isinstance(value, source.model) and source.remote_field:
                source = source.remote_field.model._meta.get_field(source.remote_field.field_name)
            try:
                value_list.append(getattr(value, source.attname))
            except AttributeError:
                # A case like Restaurant.objects.filter(place=restaurant_instance),
                # where place is a OneToOneField and the primary key of Restaurant.
                return (value.pk,)
        return tuple(value_list)
    if not isinstance(value, tuple):
        return (value,)
    return value


class RelatedIn(In):
    def get_prep_lookup(self):
        if not isinstance(self.lhs, MultiColSource) and self.rhs_is_direct_value():
            # If we get here, we are dealing with single-column relations.
            self.rhs = [get_normalized_value(val, self.lhs)[0] for val in self.rhs]
            # We need to run the related field's get_prep_value(). Consider case
            # ForeignKey to IntegerField given value 'abc'. The ForeignKey itself
            # doesn't have validation for non-integers, so we must run validation
            # using the target field.
            if hasattr(self.lhs.output_field, 'get_path_info'):
                # Run the target field's get_prep_value. We can safely assume there is
                # only one as we don't get to the direct value branch otherwise.
                target_field = self.lhs.output_field.get_path_info()[-1].target_fields[-1]
                self.rhs = [target_field.get_prep_value(v) for v in self.rhs]
        return super().get_prep_lookup()

    def as_sql(self, compiler, connection):
        if isinstance(self.lhs, MultiColSource):
            # For multicolumn lookups we need to build a multicolumn where clause.
            # This clause is either a SubqueryConstraint (for values that need to be compiled to
            # SQL) or an OR-combined list of (col1 = val1 AND col2 = val2 AND ...) clauses.
            from django.db.models.sql.where import WhereNode, SubqueryConstraint, AND, OR

            root_constraint = WhereNode(connector=OR)
            if self.rhs_is_direct_value():
                values = [get_normalized_value(value, self.lhs) for value in self.rhs]
                for value in values:
                    value_constraint = WhereNode()
                    for source, target, val in zip(self.lhs.sources, self.lhs.targets, value):
                        lookup_class = target.get_lookup('exact')
                        lookup = lookup_class(target.get_col(self.lhs.alias, source), val)
                        value_constraint.add(lookup, AND)
                    root_constraint.add(value_constraint, OR)
            else:
                root_constraint.add(
                    SubqueryConstraint(
                        self.lhs.alias, [target.column for target in self.lhs.targets],
                        [source.name for source in self.lhs.sources], self.rhs),
                    AND)
            return root_constraint.as_sql(compiler, connection)
        else:
            if (not getattr(self.rhs, 'has_select_fields', True) and
                    not getattr(self.lhs.field.target_field, 'primary_key', False)):
                self.rhs.clear_select_clause()
                if (getattr(self.lhs.output_field, 'primary_key', False) and
                        self.lhs.output_field.model == self.rhs.model):
                    # A case like Restaurant.objects.filter(place__in=restaurant_qs),
                    # where place is a OneToOneField and the primary key of
                    # Restaurant.
                    target_field = self.lhs.field.name
                else:
                    target_field = self.lhs.field.target_field.name
                self.rhs.add_fields([target_field], True)
            return super().as_sql(compiler, connection)


class RelatedLookupMixin:
    def get_prep_lookup(self):
        if not isinstance(self.lhs, MultiColSource) and self.rhs_is_direct_value():
            # If we get here, we are dealing with single-column relations.
            self.rhs = get_normalized_value(self.rhs, self.lhs)[0]
            # We need to run the related field's get_prep_value(). Consider case
            # ForeignKey to IntegerField given value 'abc'. The ForeignKey itself
            # doesn't have validation for non-integers, so we must run validation
            # using the target field.
            if self.prepare_rhs and hasattr(self.lhs.output_field, 'get_path_info'):
                # Get the target field. We can safely assume there is only one
                # as we don't get to the direct value branch otherwise.
                target_field = self.lhs.output_field.get_path_info()[-1].target_fields[-1]
                self.rhs = target_field.get_prep_value(self.rhs)

        return super().get_prep_lookup()

    def as_sql(self, compiler, connection):
        if isinstance(self.lhs, MultiColSource):
            assert self.rhs_is_direct_value()
            self.rhs = get_normalized_value(self.rhs, self.lhs)
            from django.db.models.sql.where import WhereNode, AND
            root_constraint = WhereNode()
            for target, source, val in zip(self.lhs.targets, self.lhs.sources, self.rhs):
                lookup_class = target.get_lookup(self.lookup_name)
                root_constraint.add(
                    lookup_class(target.get_col(self.lhs.alias, source), val), AND)
            return root_constraint.as_sql(compiler, connection)
        return super().as_sql(compiler, connection)


class RelatedExact(RelatedLookupMixin, Exact):
    pass


class RelatedLessThan(RelatedLookupMixin, LessThan):
    pass


class RelatedGreaterThan(RelatedLookupMixin, GreaterThan):
    pass


class RelatedGreaterThanOrEqual(RelatedLookupMixin, GreaterThanOrEqual):
    pass


class RelatedLessThanOrEqual(RelatedLookupMixin, LessThanOrEqual):
    pass


class RelatedIsNull(RelatedLookupMixin, IsNull):
    pass



from django.db import migrations, models

import scarlet.scheduling.fields


class Migration(migrations.Migration):

    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
    ]

    operations = [
        migrations.CreateModel(
            name="Schedule",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("object_args", scarlet.scheduling.fields.JSONField()),
                ("when", models.DateTimeField()),
                ("action", models.CharField(max_length=255, null=True)),
                ("json_args", scarlet.scheduling.fields.JSONField()),
                (
                    "content_type",
                    models.ForeignKey(
                        to="contenttypes.ContentType",
                        on_delete=models.deletion.CASCADE,
                    ),
                ),
            ],
        ),
    ]



import json

from django import http
from django.contrib import messages
from django.core.serializers.json import DjangoJSONEncoder
from django.shortcuts import render
from django.template.defaultfilters import slugify
from django.template.loader import render_to_string
from django.utils.encoding import force_text


class RenderResponse:
    """
    Render a template. Doesn't do anything special with css/js
    as per current front end direction.

    :param template: The template to render.
    :param partial_base: The template to use as a base for \
    partial rendering. IE: ajax requests.
    :param base: The template to use a the base template.
    """

    template = None
    base = "base.html"
    partial_base = "partial.html"

    def __init__(self, **kwargs):
        # Go through keyword arguments and save to instance
        for key, value in kwargs.items():
            setattr(self, key, value)

    def update_kwargs(self, request, **kwargs):
        """
        Hook for adding data to the context before
        rendering a template.

        :param kwargs: The current context keyword arguments.
        :param request: The current request object.
        """
        if "base" not in kwargs:
            kwargs["base"] = self.base
            if request.is_ajax() or request.GET.get("json"):
                kwargs["base"] = self.partial_base

        return kwargs

    def render(self, request, redirect_url=None, **kwargs):
        """
        Uses `self.template` to render a response.

        :param request: The current request object.
        :param redirect_url: If given this will return the \
        redirect method instead of rendering the normal template. \
        Renders providing this argument are referred to as a \
        'render redirect' in this documentation.
        :param kwargs: The current context keyword arguments.
        """
        if redirect_url:
            # Redirection is used when we click on `Save` for ordering
            # items on `ListView`. `kwargs` contains `message` but that
            # one is not passing through redirection. That's the reason for using
            # directly `messages` and get message on result template
            if kwargs.get("message"):
                messages.success(request, kwargs.get("message"))
            return self.redirect(request, redirect_url, **kwargs)

        kwargs = self.update_kwargs(request, **kwargs)
        return render(request, self.template, kwargs)

    def redirect(self, request, url, renderer=None, **kwargs):
        """
        Hook for changing redirect behavior. Should
        return a HttpResponse object. Default implementation
        redirects to the given url.

        :param request: The current request object.
        :param url: The url to redirect to.
        :param kwargs: The current context keyword arguments.
        """
        return http.HttpResponseRedirect(url)


class CMSRender(RenderResponse):
    """
    Render a template to use in the cms application. Inherits
    from RenderResponse. Used by most CMS views.
    """

    def update_kwargs(self, request, **kwargs):
        """
        Adds variables to the context that are expected by the
        base cms templates.

        * **navigation** - The side navigation for this bundle and user.
        * **dashboard** - The list of dashboard links for this user.
        * **object_header** - If no 'object_header' was passed in the \
        current context and the current bundle is set to get it's \
        object_header from it's parent, this will get that view and render \
        it as a string. Otherwise 'object_header will remain unset.
        * **subitem** - This is set to true if we rendered a new object_header \
        and the object used to render that string is not present in the \
        context args as 'obj'. This effects navigation and wording in the \
        templates.
        """

        kwargs = super().update_kwargs(request, **kwargs)

        # Check if we need to to include a separate object
        # bundle for the title
        bundle = kwargs.get("bundle")
        url_kwargs = kwargs.get("url_params")
        view = None
        if bundle:
            view, name = bundle.get_object_header_view(
                request, url_kwargs, parent_only=True
            )

        kwargs["dashboard"] = bundle.admin_site.get_dashboard_urls(request)

        if view:
            obj = view.get_object()
            if "object_header" not in kwargs:
                kwargs["object_header"] = bundle._render_view_as_string(
                    view, name, request, url_kwargs
                )
            if obj and obj != kwargs.get("obj"):
                kwargs["subitem"] = True
        return kwargs


class ChoicesRender:
    """
    A Renderer meant to render an object list view as JSON.
    Used by ListView when called with ?type=choices.
    """

    def get_different_page(self, request, page):
        """
        Returns a url that preserves the current querystring
        while changing the page requested to `page`.
        """

        if page:
            qs = request.GET.copy()
            qs["page"] = page
            return f"{request.path_info}?{qs.urlencode()}"
        return None

    def get_label_attr(self, label):
        attr = label.attr
        if label.attr == "__str__":
            attr = force_text(slugify(label.name))
        if hasattr(attr, "__call__"):
            attr = attr.__name__
        return attr

    def get_object_list(self, adm_list):
        l = []
        labels = list(adm_list.labels())
        for row in adm_list:
            data = {
                "id": row.instance.pk,
            }

            for label in labels:
                d = row.get_value(label.attr, 1)
                if callable(d):
                    d = d()
                data[self.get_label_attr(label)] = force_text(d)
            l.append(data)
        return l

    def get_fields(self, adm_list):
        data = {}
        for label in adm_list.labels():
            data[self.get_label_attr(label)] = {
                "name": force_text(label.name),
                "sortable": label.sortable,
                "order_type": label.order_type,
            }
        return data

    def render(self, request, **kwargs):
        """
        Returns a JSON representation of a objects list page.
        The json has the following attributes:

        * **is_paginated** - Is the list paginated.
        * **results** - A list of objects, where each object has an \
        attribute/value for each field in the list. An 'id' attribute \
        is always included.
        * **fields** - An object who's properties are the fields \
        in the results list. Each property will have an object with \
        the the following attributes:
            * **name** - The verbose name of the field.
            * **sortable** - Can this column be sorted. True or False.
            * **order_type** - What is the current order of this column.

        The following attributes only appear if the list is paginated:

        * **count** - If the list is paginated, how many objects \
        total are there.
        * **page** - Current page number.
        * **next** - The full link to the next page.
        * **previous** - The full link to the previous page.

        If the list can be filtered the following attribute is included:

        * **params** - An object who's properties are the filter options. \
            Each property contains an object with the following attributes:
            * **value** - If the current result list has been filtered by \
            this field then value will contain the filter value that was used.
            * **choices** - If the field is a choice field this will contain \
            the options.

        Example JSON:

        ::

            {"count": 1,
            "fields": {
                "name": {"sortable": true, "name": "name", "order_type": "asc"}
            },
            "results": [{"id": 12, "name": "Test"}],
            "next": "",
            "params": {"name": {"value": null}},
            "is_paginated": true,
            "page": 1,
            "previous": ""}
        """
        data = {"is_paginated": kwargs.get("is_paginated")}

        if data.get("is_paginated"):
            page = kwargs["page_obj"]

            next_p = ""
            previous = ""
            if page.has_next():
                next_p = self.get_different_page(request, page.number + 1)

            if page.has_previous():
                previous = self.get_different_page(request, page.number - 1)

            data.update(
                {
                    "count": page.paginator.count,
                    "page": page.number,
                    "next": next_p,
                    "previous": previous,
                }
            )

        if kwargs.get("filter_form"):
            exclude = request.GET.getlist("exclude")
            filter_form = {}
            form = kwargs.get("filter_form")
            for name in form.get_search_fields(exclude):
                k = form[name]
                obj = {}
                obj["value"] = k.value()
                obj["label"] = k.label
                if hasattr(k.field, "choices"):
                    obj["choices"] = k.field.choices

                filter_form[k.name] = obj

            data["params"] = filter_form

        adm_list = kwargs["list"]
        data["fields"] = self.get_fields(adm_list)
        data["results"] = self.get_object_list(adm_list)
        return http.HttpResponse(json.dumps(data, cls=DjangoJSONEncoder))


class RenderString(RenderResponse):
    """
    A Renderer that returns a rendered string instead of
    a HttpResponse object. Inherits from RenderResponse.
    Used by CMS views when called with render_type = 'string'.
    """

    def render(self, request, **kwargs):
        kwargs = self.update_kwargs(request, **kwargs)
        return render_to_string(self.template, kwargs, request)


class PopupRender(RenderResponse):
    """
    A Renderer that forces a special popup base template to be
    used. Returns a rendered response when a redirect is requested
    instead of redirecting to the given url. Used by FormView when called
    with ?popup=1.

    :param base: The popup only base template.
    :param redirect_template: The template to use for redirect renders.
    """

    base = "cms/base_popup.html"
    redirect_template = "cms/popup_redirect.html"

    def update_kwargs(self, request, **kwargs):
        kwargs["base"] = self.base
        return kwargs

    def redirect(self, request, url, **kwargs):
        return render(request, self.redirect_template, kwargs)



from django.db.models import Prefetch
from rest_framework.generics import ListAPIView, RetrieveAPIView
from rest_framework.renderers import JSONRenderer
from scarlet.pagebuilder.serializers import PageListSerializer, PageSerializer
from . import get_page_model

Page = get_page_model()


class PageListView(ListAPIView):
    queryset = Page.objects.all()
    serializer_class = PageSerializer
    renderer_classes = [JSONRenderer]
    pagination_class = None


class PageDetailsView(RetrieveAPIView):
    serializer_class = PageSerializer
    queryset = Page.objects.all()
    renderer_classes = [JSONRenderer]
    lookup_field = "slug"

    def get_queryset(self):
        return (
            super()
            .get_queryset()
            .prefetch_related(
                "page_hero_modules",
                "page_two_column_modules",
                "page_icon_list_modules",
                "page_header_modules",
                "page_faq_modules",
                "page_location_modules",
                "page_gallery_modules",
                "page_carousel_modules",
            )
        )



from .. import Provider as CompanyProvider


class Provider(CompanyProvider):
    formats = (
        '{{last_name}} {{company_suffix}}',
        '{{last_name}} {{last_name}} {{company_suffix}}',
        '{{last_name}}',
    )

    company_suffixes = (
        'd.o.o.', 'd.d.', 'j.d.o.o.',
    )



from MySQLdb.constants import FIELD_TYPE

from django.contrib.gis.gdal import OGRGeomType
from django.db.backends.mysql.introspection import DatabaseIntrospection


class MySQLIntrospection(DatabaseIntrospection):
    # Updating the data_types_reverse dictionary with the appropriate
    # type for Geometry fields.
    data_types_reverse = DatabaseIntrospection.data_types_reverse.copy()
    data_types_reverse[FIELD_TYPE.GEOMETRY] = 'GeometryField'

    def get_geometry_type(self, table_name, geo_col):
        with self.connection.cursor() as cursor:
            # In order to get the specific geometry type of the field,
            # we introspect on the table definition using `DESCRIBE`.
            cursor.execute('DESCRIBE %s' %
                           self.connection.ops.quote_name(table_name))
            # Increment over description info until we get to the geometry
            # column.
            for column, typ, null, key, default, extra in cursor.fetchall():
                if column == geo_col:
                    # Using OGRGeomType to convert from OGC name to Django field.
                    # MySQL does not support 3D or SRIDs, so the field params
                    # are empty.
                    field_type = OGRGeomType(typ).django
                    field_params = {}
                    break
        return field_type, field_params

    def supports_spatial_index(self, cursor, table_name):
        # Supported with MyISAM/Aria, or InnoDB on MySQL 5.7.5+/MariaDB 10.2.2+
        storage_engine = self.get_storage_engine(cursor, table_name)
        if storage_engine == 'InnoDB':
            return self.connection.mysql_version >= (
                (10, 2, 2) if self.connection.mysql_is_mariadb else (5, 7, 5)
            )
        return storage_engine in ('MyISAM', 'Aria')
